#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:42 下午
# @Author  : Han Yu
# @File    : SU2_rules.py

"""
the file describe transform rules that decompose SU(2) into instruction set.
"""

import numpy as np
from numpy import arccos, linalg

from QuICT.core.gate import *


def _arccos(value):
    """ calculate arccos(value)

    Args:
        value: the cos value

    Returns:
        float: the corresponding angle
    """
    if value < -1:
        value = -1
    elif value > 1:
        value = 1
    return arccos(value)


def _check2pi(theta, eps=1e-15):
    """ check whether theta is a multiple of 2π

    Args:
        theta(float): the angle to be checked
        eps(float): tolerate error

    Returns:
        bool: whether theta is a multiple of 2π
    """
    multiple = np.round(theta / (2 * np.pi))
    return abs(2 * np.pi * multiple - theta) < eps


def zyz_rule(gate):
    """ decomposition the unitary gate with 2 * 2 unitary into Rz Ry Rz sequence

    Args:
        gate(Unitary): the gate to be decomposed

    Returns:
        compositeGate: a list of compositeGate
    """
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-13

    det = linalg.det(unitary)
    beta_plus_delta = 0
    beta_dec_delta = 0
    if abs(det - 1) > eps:
        unitary[:] /= np.sqrt(det)

    if abs(unitary[0, 0]) > abs(unitary[0, 1]) > eps:
        gamma = _arccos((2 * (unitary[0, 0] * unitary[1, 1]).real - 1))
    else:
        gamma = _arccos((2 * (unitary[0, 1] * unitary[1, 0]).real + 1))
    if abs(unitary[0, 0]) > eps:
        beta_plus_delta = -np.angle(unitary[0, 0] / np.cos(gamma / 2)) * 2
    if abs(unitary[0, 1]) > eps:
        beta_dec_delta = np.angle(unitary[1, 0] / np.sin(gamma / 2)) * 2

    beta = (beta_plus_delta + beta_dec_delta) / 2
    delta = beta_plus_delta - beta
    gates = CompositeGate()
    with gates:
        if not _check2pi(delta):
            Rz(delta) & targ
        if not _check2pi(gamma):
            Ry(gamma) & targ
        if not _check2pi(beta):
            Rz(beta) & targ
    return gates


def xyx_rule(gate):
    """ decomposition the unitary gate with 2 * 2 unitary into Rx Ry Rx sequence

    Args:
        gate(Unitary): the gate to be decomposed

    Returns:
        compositeGate: a list of compositeGate
    """
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-13
    unitary = np.array([
        [0.5 * (unitary[0, 0] + unitary[0, 1] + unitary[1, 0] + unitary[1, 1]),
         0.5 * (unitary[0, 0] - unitary[0, 1] + unitary[1, 0] - unitary[1, 1])],
        [0.5 * (unitary[0, 0] + unitary[0, 1] - unitary[1, 0] - unitary[1, 1]),
         0.5 * (unitary[0, 0] - unitary[0, 1] - unitary[1, 0] + unitary[1, 1])]
    ], dtype=np.complex128)
    det = linalg.det(unitary)
    beta_plus_delta = 0
    beta_dec_delta = 0
    if abs(det - 1) > eps:
        unitary[:] /= np.sqrt(det)

    if abs(unitary[0, 0]) > abs(unitary[0, 1]) > eps:
        gamma = _arccos((2 * (unitary[0, 0] * unitary[1, 1]).real - 1))
    else:
        gamma = _arccos((2 * (unitary[0, 1] * unitary[1, 0]).real + 1))
    if abs(unitary[0, 0]) > eps:
        beta_plus_delta = -np.angle(unitary[0, 0] / np.cos(gamma / 2)) * 2
    if abs(unitary[0, 1]) > eps:
        beta_dec_delta = np.angle(unitary[1, 0] / np.sin(gamma / 2)) * 2

    beta = (beta_plus_delta + beta_dec_delta) / 2
    delta = beta_plus_delta - beta
    gates = CompositeGate()
    with gates:
        if not _check2pi(delta):
            Rx(delta) & targ
        if not _check2pi(gamma):
            Ry(-gamma) & targ
        if not _check2pi(beta):
            Rx(beta) & targ
    return gates


def ibmq_rule(gate):
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-13

    det = linalg.det(unitary)
    beta_plus_delta = 0
    beta_dec_delta = 0
    if abs(det - 1) > eps:
        unitary[:] /= np.sqrt(det)

    if abs(unitary[0, 0]) > abs(unitary[0, 1]) > eps:
        gamma = _arccos((2 * (unitary[0, 0] * unitary[1, 1]).real - 1))
    else:
        gamma = _arccos((2 * (unitary[0, 1] * unitary[1, 0]).real + 1))
    if abs(unitary[0, 0]) > eps:
        beta_plus_delta = -np.angle(unitary[0, 0] / np.cos(gamma / 2)) * 2
    if abs(unitary[0, 1]) > eps:
        beta_dec_delta = np.angle(unitary[1, 0] / np.sin(gamma / 2)) * 2

    beta = (beta_plus_delta + beta_dec_delta) / 2
    delta = beta_plus_delta - beta
    gates = CompositeGate()
    with gates:
        if not _check2pi(delta):
            Rz(delta) & targ
        if not _check2pi(gamma):
            SX & targ
            Rz(gamma) & targ
            SX & targ
            X & targ
        if not _check2pi(beta):
            Rz(beta) & targ
    return gates


def zxz_rule(gate):
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-13

    det = linalg.det(unitary)
    beta_plus_delta = 0
    beta_dec_delta = 0
    if abs(det - 1) > eps:
        unitary[:] /= np.sqrt(det)

    beta_plus_delta = 2 * np.angle(unitary[1, 1])
    beta_dec_delta = 2 * np.angle(unitary[1, 0]) + np.pi
    gamma = 2 * _arccos(np.abs(unitary[0, 0]))
    beta = (beta_plus_delta + beta_dec_delta) / 2
    delta = beta_plus_delta - beta

    gates = CompositeGate()
    with gates:
        if not _check2pi(delta):
            Rz(delta) & targ
        if not _check2pi(gamma):
            Rx(gamma) & targ
        if not _check2pi(beta):
            Rz(beta) & targ
    return gates


def hrz_rule(gate):
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-13

    det = linalg.det(unitary)
    beta_plus_delta = 0
    beta_dec_delta = 0
    if abs(det - 1) > eps:
        unitary[:] /= np.sqrt(det)

    beta_plus_delta = 2 * np.angle(unitary[1, 1])
    beta_dec_delta = 2 * np.angle(unitary[1, 0]) + np.pi
    gamma = 2 * _arccos(np.abs(unitary[0, 0]))
    beta = (beta_plus_delta + beta_dec_delta) / 2
    delta = beta_plus_delta - beta

    gates = CompositeGate()
    with gates:
        if not _check2pi(delta):
            Rz(delta) & targ
        if not _check2pi(gamma):
            H & targ
            Rz(gamma) & targ
            H & targ
        if not _check2pi(beta):
            Rz(beta) & targ
    return gates


def u3_rule(gate):
    unitary = gate.matrix
    targ = gate.targ
    eps = 1e-6

    # u3[0, 0] is real
    z = np.exp(1j * np.angle(unitary[0, 0]))
    unitary = unitary / z

    theta = np.arccos(unitary[0, 0])
    sint = np.sin(theta)
    if abs(sint) >= eps:
        lamda = np.angle(unitary[0, 1] / -sint)
        phi = np.angle(unitary[1, 0] / sint)
    else:
        lamda = 0
        phi = np.angle(unitary[1, 1] / np.cos(theta))

    if _check2pi(theta, eps):
        theta = 0
    if _check2pi(lamda, eps):
        lamda = 0
    if _check2pi(phi, eps):
        phi = 0
    g = build_gate(GateType.u3, targ, [theta * 2, phi, lamda])
    gates = CompositeGate(gates=[g])
    return gates



from functools import wraps

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import CompositeGate


class OutputAligner(object):
    """
    Decorating class that keeps the type of output aligned with that of input for QCDA execute functions

    Valid type of output are restricted to be CompositeGate or Circuit.
    For functions with other types of input, the type of output could be assigned.
    """
    def __init__(self, output=None):
        """
        Args:
            output(class, optional): assigned type of output in [CompositeGate, Circuit]
        """
        if output is not None:
            assert output in [CompositeGate, Circuit], ValueError('Invalid output type')
        self.output = output

    def __call__(self, func):
        """
        Args:
            func(callable): function to be decorated

        Returns:
            callable: func with whose output type aligned
        """

        @wraps(func)
        def align_func(object, input):
            """
            For QCDA execute functions, their input would be restricted to be one argument representing
            the input that would be synthesized, optimized or mapped, while other parameters needed in
            the function would be passed by the __init__ method.

            Args:
                object(Object): the object that func belongs to
                input: input of func, the following types are supported.
                    1. CompositeGate: the input is a CompositeGate
                    2. Circuit: the input is a Circuit
                    3. numpy.ndarray: the input is a unitary matrix

            Returns:
                CompositeGate/Circuit: output of func with whose output type aligned
            """
            # Record width of input
            if isinstance(input, CompositeGate) or isinstance(input, Circuit):
                self.output = type(input)
                width = input.width()
            elif isinstance(input, np.ndarray):
                assert input.ndim == 2 and input.shape[0] == input.shape[1],\
                    ValueError('Input is not a square matrix')
                width = int(round(np.log2(input.shape[0])))
                assert 1 << width == input.shape[0], ValueError('Input is not a 2^n * 2^n matrix')
            else:
                raise TypeError('Invalid input type')

            assert self.output in [CompositeGate, Circuit], TypeError('Invalid output type')

            output = func(object, input)
            if isinstance(output, self.output):
                return output
            if self.output == CompositeGate:
                output: Circuit
                gates = CompositeGate(gates=output.gates)
                return gates
            if self.output == Circuit:
                output: CompositeGate
                circuit = Circuit(width)
                circuit.extend(output)
                return circuit

        return align_func



import os
import json


def options_validation(options: dict, device: str, backend: str = None):
    """ Validation the given options is related simulator's parameters.

    Args:
        options (dict): The given simulator's parameters.
        device (str): The device of the simulator. One of [CPU, GPU]
        backend (str): The backend for the simulator. One of [unitary, state_vector, density_matrix]
    """
    # Load simulator's options dict
    data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "simulator_parameters.json")
    with open(data_path, "r") as f:
        simulator_options = json.load(f)

    assert device in ["CPU", "GPU"], "Device should be one of [CPU, GPU]."
    simulator_options = simulator_options[device]
    # Select target simulator's option by input
    simulator_options_keys = []
    if backend is not None:
        assert backend in ["unitary", "state_vector", "density_matrix"], \
            "backend should be one of [unitary, state_vector, density_matrix]."

        simulator_options_keys.append(set(simulator_options[backend].keys()))
    else:
        for backend, parameters in simulator_options.items():
            simulator_options_keys.append(set(parameters.keys()))

    # Validation the input options
    input_option_keys = set(options.keys())
    for parameters_set in simulator_options_keys:
        union_params = input_option_keys & parameters_set
        if len(union_params) == len(input_option_keys):
            return True

    return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# Modification Notice: Code revised for QuICT

import time
from collections import Counter

from QuICT.core.circuit import Circuit
from QuICT.core.gate import *


class TemplateSearching:
    """ for clifford gates search for S, CNOT, H gates """

    def __init__(self, qubit_num, gate_num, gate_dep):
        self.qubit_num = qubit_num
        self.gate_num = gate_num
        self.dep = gate_dep
        self.template_list = []
        self.temp_qubit_dep = [0] * qubit_num
        self.target = []
        for i in range(qubit_num):
            self.target.append(i)

    def copy_circuit(self, temp_circuit):
        new_circuit = Circuit(temp_circuit.width())
        new_circuit.extend(temp_circuit.gates)
        return new_circuit

    def identity(self, temp_circuit):
        matrix = temp_circuit.matrix()
        n = np.size(matrix, 0)
        tot = 0
        for i in range(n):
            for j in range(n):
                if i != j:
                    tot += abs(matrix[i, j])
                else:
                    tot += abs(matrix[i, j] - 1)
        return abs(tot) < 1e-2

    def check_minimum(self, temp_circuit):
        n = temp_circuit.size()
        if self.identity(temp_circuit):
            for i in range(n):
                for j in range(1, n - i):
                    new_circuit = self.copy_circuit(temp_circuit.sub_circuit(i, j, self.target))
                    if self.identity(new_circuit):
                        return False
            return True
        return False

    def graph_isomorphism(self, inform_list_a, inform_list_b, status_list_a, status_list_b):

        # check if an ordered-list graph is not isomorphism with the other
        flag_all_mapping = True
        for vertex_a in range(self.qubit_num):
            if self.mapping[vertex_a] == -1:
                flag_all_mapping = False
                break

        if flag_all_mapping:
            flag_check_graph = True
            for i in range(self.qubit_num):
                j = self.mapping[i]
                if len(inform_list_a[i]) == len(inform_list_b[j]) and len(status_list_a[i]) == len(status_list_b[j]):
                    for k in range(len(inform_list_a[i])):
                        if not flag_check_graph:
                            break
                        if status_list_a[i][k] != status_list_b[j][k]:
                            flag_check_graph = False
                        else:
                            if status_list_a[i][k] == 'S':
                                flag_check_graph = flag_check_graph & (
                                    Counter(inform_list_a[i][k]) == Counter(inform_list_b[j][k])
                                )
                            else:
                                mapping_list_a = Counter(inform_list_a[i][k])
                                mapping_list_b = Counter(inform_list_b[j][k])
                                if len(mapping_list_a) == len(mapping_list_b):
                                    for p in mapping_list_a:
                                        q = self.mapping[p]
                                        flag_check_graph = flag_check_graph & (mapping_list_a[p] == mapping_list_b[q])
                                else:
                                    flag_check_graph = False
                                    break
                else:
                    flag_check_graph = False
                    break
            return flag_check_graph
        else:
            flag_check_graph = False
            for i in range(self.qubit_num):
                if not self.mapped[i]:
                    if (
                        len(inform_list_a[vertex_a]) == len(inform_list_b[i]) and
                        len(status_list_a[vertex_a]) == len(status_list_b[i])
                    ):
                        flag_check_list = True
                        self.mapped[i] = True
                        self.mapping[vertex_a] = i
                        for j in range(len(status_list_a[vertex_a])):
                            check_a = status_list_a[vertex_a][j]
                            check_b = status_list_b[i][j]
                            flag_check_list = flag_check_list & (check_a == check_b)
                        if flag_check_list:
                            flag_check_graph = flag_check_graph or self.graph_isomorphism(
                                inform_list_a, inform_list_b, status_list_a, status_list_b
                            )
                            if flag_check_graph:
                                break
                        if not flag_check_graph:
                            self.mapped[i] = False
                            self.mapping[vertex_a] = -1

            return flag_check_graph

    def commutative_processing(self, temp_circuit):
        for i in range(len(temp_circuit.gates) - 1):
            gate_a = temp_circuit.gates[i]
            gate_b = temp_circuit.gates[i + 1]
            if gate_a.commutative(gate_b):
                if gate_a.is_control_single() and not gate_b.is_control_single():
                    temp_circuit.gates[i] = gate_b
                    temp_circuit.gates[i + 1] = gate_a
                    temp_circuit = self.commutative_processing(temp_circuit)
        return temp_circuit

    def check_circuit_not_isomorphism(self, circuit_a, circuit_b):

        # check if a circuit is not isomorphism with the other
        # first only check cnot circuit

        # commutative processing
        circuit_a = self.commutative_processing(circuit_a)
        circuit_b = self.commutative_processing(circuit_b)

        inform_list_a = []
        status_list_a = []
        inform_list_b = []
        status_list_b = []
        for i in range(self.qubit_num):
            inform_list_a.append([])
            status_list_a.append([])
            inform_list_b.append([])
            status_list_b.append([])

        temp_inform_list = []
        for i in range(self.qubit_num):
            temp_inform_list.append([])
        for gate_a in circuit_a.gates:
            if gate_a.is_control_single():
                temp_gate_c = gate_a.carg
                temp_gate_t = gate_a.targ
                if status_list_a[temp_gate_c] == []:
                    status_list_a[temp_gate_c].append('C')
                    temp_inform_list[temp_gate_c].append(temp_gate_t)
                else:
                    if status_list_a[temp_gate_c][-1] == 'C':
                        temp_inform_list[temp_gate_c].append(temp_gate_t)
                    else:
                        inform_list_a[temp_gate_c].append(temp_inform_list[temp_gate_c])
                        status_list_a[temp_gate_c].append('C')
                        temp_inform_list[temp_gate_c] = [temp_gate_t]
            else:
                temp_gate = gate_a.targ
                if status_list_a[temp_gate] == []:
                    status_list_a[temp_gate].append('S')
                    if gate_a.type == GateType.h:
                        temp_inform_list[temp_gate].append(1)
                    else:
                        temp_inform_list[temp_gate].append(2)
                else:
                    if status_list_a[temp_gate][-1] == 'S':
                        if gate_a.type == GateType.h:
                            temp_inform_list[temp_gate].append(1)
                        else:
                            temp_inform_list[temp_gate].append(2)
                    else:
                        inform_list_a[temp_gate].append(temp_inform_list[temp_gate])
                        status_list_a[temp_gate].append('S')
                        if gate_a.type == GateType.h:
                            temp_inform_list[temp_gate] = [1]
                        else:
                            temp_inform_list[temp_gate] = [2]

        for i in range(self.qubit_num):
            if temp_inform_list[i] != []:
                inform_list_a[i].append(temp_inform_list[i])

        temp_inform_list = []
        for i in range(self.qubit_num):
            temp_inform_list.append([])
        for gate_b in circuit_b.gates:
            if gate_b.is_control_single():
                temp_gate_c = gate_b.carg
                temp_gate_t = gate_b.targ
                if status_list_b[temp_gate_c] == []:
                    status_list_b[temp_gate_c].append('C')
                    temp_inform_list[temp_gate_c].append(temp_gate_t)
                else:
                    if status_list_b[temp_gate_c][-1] == 'C':
                        temp_inform_list[temp_gate_c].append(temp_gate_t)
                    else:
                        inform_list_b[temp_gate_c].append(temp_inform_list[temp_gate_c])
                        status_list_b[temp_gate_c].append('C')
                        temp_inform_list[temp_gate_c] = [temp_gate_t]
            else:
                temp_gate = gate_b.targ
                if status_list_b[temp_gate] == []:
                    status_list_b[temp_gate].append('S')
                    if gate_b.type == GateType.h:
                        temp_inform_list[temp_gate].append(1)
                    else:
                        temp_inform_list[temp_gate].append(2)
                else:
                    if status_list_b[temp_gate][-1] == 'S':
                        if gate_b.type == GateType.h:
                            temp_inform_list[temp_gate].append(1)
                        else:
                            temp_inform_list[temp_gate].append(2)
                    else:
                        inform_list_b[temp_gate].append(temp_inform_list[temp_gate])
                        status_list_b[temp_gate].append('S')
                        if gate_b.type == GateType.h:
                            temp_inform_list[temp_gate] = [1]
                        else:
                            temp_inform_list[temp_gate] = [2]

        for i in range(self.qubit_num):
            if temp_inform_list[i] != []:
                inform_list_b[i].append(temp_inform_list[i])

        self.mapping = []
        self.mapped = []
        for i in range(self.qubit_num):
            self.mapping.append(-1)
            self.mapped.append(False)
        flag_carg_graph = self.graph_isomorphism(inform_list_a, inform_list_b, status_list_a, status_list_b)

        return not flag_carg_graph

#    def check_list_not_isomorphism(self, temp_circuit):

        # check if the circuit is not isomorphism with every template in the list

#        no_isomorphism = True
#        for template in self.template_list:
#            no_isomorphism = no_isomorphism & self.check_circuit_not_isomorphism(temp_circuit, template)
#        return no_isomorphism

    def search(self, temp_gate_num, temp_circuit):

        # check whether it is a template
        if 1 <= temp_gate_num <= self.gate_num:
            if self.identity(temp_circuit):
                # if self.check_list_not_isomorphism(temp_circuit):
                #     self.template_list.append(temp_circuit)
                self.template_list.append([temp_circuit, self.check_minimum(temp_circuit)])
                return
        if temp_gate_num == self.gate_num:
            return

        # brute-force searching
        # s gate
        for s in range(0, self.qubit_num):
            if self.temp_qubit_dep[s] + 1 <= self.dep:
                new_circuit = self.copy_circuit(temp_circuit)
                S | new_circuit(s)
                self.temp_qubit_dep[s] += 1
                self.search(temp_gate_num + 1, new_circuit)
                self.temp_qubit_dep[s] -= 1

        # hadamard gate
        for h in range(0, self.qubit_num):
            if self.temp_qubit_dep[h] + 1 <= self.dep:
                new_circuit = self.copy_circuit(temp_circuit)
                H | new_circuit(h)
                self.temp_qubit_dep[h] += 1
                self.search(temp_gate_num + 1, new_circuit)
                self.temp_qubit_dep[h] -= 1

        # cnot gate
        for cnot_a in range(0, self.qubit_num):
            for cnot_b in range(0, self.qubit_num):
                if cnot_a != cnot_b:
                    temp_max_dep = max(self.temp_qubit_dep[cnot_a], self.temp_qubit_dep[cnot_b]) + 1
                    if temp_max_dep <= self.dep:
                        new_circuit = self.copy_circuit(temp_circuit)
                        CX | new_circuit([cnot_a, cnot_b])
                        temp_dep_a = self.temp_qubit_dep[cnot_a]
                        temp_dep_b = self.temp_qubit_dep[cnot_b]
                        self.temp_qubit_dep[cnot_a] = temp_max_dep
                        self.temp_qubit_dep[cnot_b] = temp_max_dep
                        self.search(temp_gate_num + 1, new_circuit)
                        self.temp_qubit_dep[cnot_a] = temp_dep_a
                        self.temp_qubit_dep[cnot_b] = temp_dep_b
        return

    def run_template_searching(self):
        circuit_temp = Circuit(self.qubit_num)
        self.search(0, circuit_temp)
        len_list = len(self.template_list)
        relationship_iso = [-1] * len_list
        for i in range(len_list):
            if relationship_iso[i] == -1:
                relationship_iso[i] = i
                for j in range(len_list):
                    if relationship_iso[j] == -1 and not self.check_circuit_not_isomorphism(
                        self.template_list[i][0], self.template_list[j][0]
                    ):
                        self.template_list[i][1] = self.template_list[i][1] and self.template_list[j][1]
                        relationship_iso[j] = i
                        self.template_list[j][1] = False

        return self.template_list



from .. import InstructionSet
from ..transform_rule import ibmq_rule

from QuICT.core.gate import *

IBMQSet = InstructionSet(
    GateType.cx,
    [GateType.rz, GateType.sx, GateType.x]
)
IBMQSet.register_one_qubit_rule(ibmq_rule)



from typing import Dict, List, Tuple

from QuICT.core import *
from QuICT.core.gate import *

from .dag import DAG


class OptimizingTemplate:
    """
    Circuit template used in AutoOptimization.
    """
    def __init__(self, template: DAG, replacement: DAG = None,
                 anchor: int = 0, weight: int = 1, phase: float = 0):
        """
        Args:
            template(DAG): template circuit
            replacement(DAG): replacement circuit
            anchor(int): starting qubit of comparison with `template`
            weight(int): weight of this template
            phase(float): global phase. exp(`phase` * pi) * `replacement` == `template`
        """
        self.template = template
        self.replacement = replacement
        self.anchor = anchor
        self.weight = weight
        self.phase = phase

    def compare(self, other: Tuple[DAG.Node, int], flag_enabled=False, dummy_rz=False):
        """
        Compare the other circuit with template.
        This circuit will start from the first gate on qubit `anchor`.
        The other circuit will start from (node, wire) defined by `other`.

        Args:
            other: the start point of the other circuit
            flag_enabled(bool): Whether consider `flag` field. If true,
                nodes already with FLAG_VISITED will be
                skipped. Nodes in the matching will be set FLAG_VISITED.

        Returns:
            Dict[int, DAG.Node]: mapping from id(node of this circuit) to
                matched node in the other circuit. If not matched, return None.
        """
        return self.template.compare_circuit(other, self.anchor, flag_enabled, dummy_rz)

    def get_replacement(self, mapping):
        """
        Get a copy of replacement circuit.

        Args:
            mapping(Dict[int, DAG.Node]): the mapping found by `compare` method.
        Returns:
            DAG: a copy of replacement.
        """
        return self.replacement.copy()

    def replace(self, mapping: Dict[int, DAG.Node]):
        """
        Replace by `mapping`.

        Args:
            mapping(Dict[int, DAG.Node]): mapping from id(template node) to
                the matched node in the original circuit.
        """
        replacement = self.get_replacement(mapping)
        new_mapping = {}
        for qubit_ in range(replacement.width()):
            t_node, t_qubit = self.template.start_nodes[qubit_].successors[0]
            p_node, p_qubit = mapping[id(t_node)].predecessors[t_qubit]
            r_node = replacement.start_nodes[qubit_]
            new_mapping[id(r_node)] = (p_node, p_qubit)

            t_node, t_qubit = self.template.end_nodes[qubit_].predecessors[0]
            s_node, s_qubit = mapping[id(t_node)].successors[t_qubit]
            r_node = replacement.end_nodes[qubit_]
            new_mapping[id(r_node)] = (s_node, s_qubit)
        DAG.replace_circuit(new_mapping, replacement)

    def regrettable_replace(self, mapping: Dict[int, DAG.Node]):
        """
        Replace by `mapping`, but can undo later.

        Args:
            mapping(Dict[int, DAG.Node]): mapping from id(template node) to
                the matched node in the original circuit.
        Returns:
            Tuple[DAG, Dict[int, Tuple[DAG.Node, int]]]: (old circuit, mapping).
                Information needed to undo the replacing.
        """

        replacement = self.get_replacement(mapping)
        original = DAG(Circuit(replacement.width()),
                       build_toffoli=replacement.build_toffoli)

        new_mapping = {}
        undo_mapping = {}
        for qubit_ in range(replacement.width()):
            t_node, t_qubit = self.template.start_nodes[qubit_].successors[0]
            p_node, p_qubit = mapping[id(t_node)].predecessors[t_qubit]
            r_node = replacement.start_nodes[qubit_]
            new_mapping[id(r_node)] = (p_node, p_qubit)
            undo_mapping[id(original.start_nodes[qubit_])] = (p_node, p_qubit)

            t_node, t_qubit = self.template.end_nodes[qubit_].predecessors[0]
            s_node, s_qubit = mapping[id(t_node)].successors[t_qubit]
            r_node = replacement.end_nodes[qubit_]
            new_mapping[id(r_node)] = (s_node, s_qubit)
            undo_mapping[id(original.end_nodes[qubit_])] = (s_node, s_qubit)

        DAG.replace_circuit(new_mapping, replacement, erase_old=False)

        for qubit_ in range(replacement.width()):
            t_node, t_qubit = self.template.start_nodes[qubit_].successors[0]
            t_node = mapping[id(t_node)]
            original.start_nodes[qubit_].connect(0, t_qubit, t_node)

            t_node, t_qubit = self.template.end_nodes[qubit_].predecessors[0]
            t_node = mapping[id(t_node)]
            t_node.connect(t_qubit, 0, original.end_nodes[qubit_])

        return original, undo_mapping

    @staticmethod
    def undo_replace(original, undo_mapping):
        """
        Undo replacing.

        Args:
            original(DAG): return value of `regrettable_replace`
            undo_mapping(Dict[int, Tuple[DAG.Node, int]]): return value of `regrettable_replace`
        """
        DAG.replace_circuit(undo_mapping, original)

    def replace_all(self, dag: DAG):
        """
        Find all `template` in the `dag` and replace them with `replacement`.

        Args:
            dag(DAG): circuit to replace.
        """

        dag.reset_flag()

        matched = []
        for node in dag.topological_sort():
            mapping = self.compare((node, -1), flag_enabled=True, dummy_rz=True)
            if not mapping:
                continue
            matched.append(mapping)

        for mapping in matched:
            self.replace(mapping)

        dag.global_phase = np.mod(dag.global_phase + len(matched) * self.phase, 2 * np.pi)
        return len(matched)


class ParameterizedTemplate(OptimizingTemplate):
    """
    Circuit template with Rz gates.
    """
    def __init__(self, template: DAG, replacement: DAG = None,
                 anchor: int = 0, weight: int = 1, phase: float = 0, param_order: List[int] = None):
        super().__init__(template, replacement, anchor, weight, phase)

        self.rz_list = list(filter(lambda g: g.gate_type == GateType.rz, template.topological_sort()))
        self.param_order = list(range(len(self.rz_list))) if param_order is None else param_order

    def get_replacement(self, mapping):
        """
        Get a copy of circuit `replacement`. Set Rz phases in the copy by `param_order`.

        Args:
            mapping(Dict[int, DAG.Node]): the mapping found by `compare` method.
        Returns:
            DAG: a copy of replacement.
        """
        replacement = self.replacement.copy()
        r_rz_list = list(filter(lambda g: g.gate_type == GateType.rz, replacement.topological_sort()))
        for idx, rz in zip(self.param_order, r_rz_list):
            rz.params = mapping[id(self.rz_list[idx])].params.copy()
            if np.isclose(float(rz.params[0]), 0):
                rz.erase()
        return replacement


def get_circuit_from_list(n_qubit, gate_list):
    circ = Circuit(n_qubit)
    for Gate_, qubit_ in gate_list:
        Gate_ | circ(qubit_)
    return circ


def reverse_order(order):
    rev = [0] * len(order)
    for idx, val in enumerate(order):
        rev[val] = idx
    return rev


def generate_hadamard_gate_templates() -> List[OptimizingTemplate]:
    # FIXME weight is count of hadamard gates
    tpl_list = [
        [1, 2, 0, [[H, 0], [H, 0]], []],
        [1, 1, +np.pi / 4, [[H, 0], [S, 0], [H, 0]], [[S_dagger, 0], [H, 0], [S_dagger, 0]]],
        [1, 1, -np.pi / 4, [[H, 0], [S_dagger, 0], [H, 0]], [[S, 0], [H, 0], [S, 0]]],
        [2, 4, 0, [[H, 0], [H, 1], [CX, [0, 1]], [H, 0], [H, 1]], [[CX, [1, 0]]]],
        [2, 2, 0, [[H, 1], [S, 1], [CX, [0, 1]], [S_dagger, 1], [H, 1]], [[S_dagger, 1], [CX, [0, 1]], [S, 1]]],
        [2, 2, 0, [[H, 1], [S_dagger, 1], [CX, [0, 1]], [S, 1], [H, 1]], [[S, 1], [CX, [0, 1]], [S_dagger, 1]]],
        [1, 2, 0, [[X, 0], [X, 0]], []]
    ]

    ret = []
    for n_qubit, weight, phase, tpl, rpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        rpl_circ = get_circuit_from_list(n_qubit, rpl)
        ret.append(OptimizingTemplate(DAG(tpl_circ), DAG(rpl_circ), weight=weight, phase=phase))
    return ret


def generate_single_qubit_gate_templates() -> List[OptimizingTemplate]:
    tpl_list = [
        [2, 1, [[H, 1], [CX, [0, 1]], [H, 1]]],
        [2, 1, [[CX, [0, 1]], [Rz(0), 1], [CX, [0, 1]]]],
        [2, 0, [[CX, [0, 1]]]],
        [3, 0, [[CX, [1, 0]], [CX, [0, 2]], [CX, [1, 0]]]],
        [1, 0, [[H, 0], [X, 0], [H, 0]]]
    ]

    ret = []
    for n_qubit, anchor, tpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        ret.append(OptimizingTemplate(DAG(tpl_circ), anchor=anchor))
    return ret


def generate_cnot_targ_templates() -> List[OptimizingTemplate]:
    tpl_list = [
        [2, 1, [[CX, [0, 1]]]],
        [2, 0, [[H, 0], [CX, [0, 1]], [H, 0]]],
        [1, 0, [[X, 0]]]
    ]

    ret = []
    for n_qubit, anchor, tpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        ret.append(OptimizingTemplate(DAG(tpl_circ), anchor=anchor))
    return ret


def generate_cnot_ctrl_templates() -> List[OptimizingTemplate]:
    tpl_list = [
        [2, 0, [[CX, [0, 1]]]],
        [1, 0, [[Rz(0), 0]]]
    ]

    ret = []
    for n_qubit, anchor, tpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        ret.append(OptimizingTemplate(DAG(tpl_circ), anchor=anchor))
    return ret


def generate_gate_preserving_rewrite_template() -> List[ParameterizedTemplate]:
    tpl_list = [
        [3, [2, 1, 0],
         [[CX, [0, 2]], [Rz(1), 2], [CX, [1, 2]], [Rz(2), 2], [CX, [0, 2]], [Rz(3), 2], [CX, [1, 2]]],
         [[CX, [1, 2]], [Rz(3), 2], [CX, [0, 2]], [Rz(2), 2], [CX, [1, 2]], [Rz(1), 2], [CX, [0, 2]]]
         ],
        [3, [1, 0],
         [[CX, [0, 2]], [Rz(1), 2], [CX, [1, 2]], [Rz(2), 2], [CX, [1, 2]]],
         [[CX, [1, 2]], [CX, [0, 2]], [Rz(2), 2], [CX, [1, 2]], [Rz(1), 2]]
         ]
    ]

    ret = []
    for n_qubit, order, tpl, rpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        rpl_circ = get_circuit_from_list(n_qubit, rpl)
        ret.append(ParameterizedTemplate(DAG(tpl_circ), DAG(rpl_circ), param_order=order))
        ret.append(ParameterizedTemplate(DAG(rpl_circ), DAG(tpl_circ), param_order=reverse_order(order)))
        return ret


def generate_gate_reducing_rewrite_template() -> List[ParameterizedTemplate]:
    tpl_list = [
        [4, 2, [2, 1, 0, 3],
         [[CX, [2, 1]], [CX, [0, 3]], [CX, [1, 3]], [Rz(1), 3], [CX, [1, 3]], [CX, [2, 3]],
          [Rz(2), 3], [CX, [0, 3]], [Rz(3), 3], [CX, [1, 3]], [Rz(4), 3], [CX, [2, 3]], [CX, [1, 3]]],
         [[CX, [2, 3]], [Rz(3), 3], [CX, [0, 3]], [Rz(2), 3], [CX, [1, 3]], [Rz(1), 3],
          [CX, [0, 3]], [CX, [2, 3]], [Rz(4), 3], [CX, [1, 3]], [CX, [2, 1]]]
         ],
        [4, 2, [0, 2, 4, 3, 1],
         [[CX, [2, 1]], [Rz(1), 3], [CX, [2, 3]], [Rz(2), 3], [CX, [1, 3]], [Rz(3), 3], [CX, [1, 3]],
          [CX, [0, 3]], [Rz(4), 3], [CX, [2, 3]], [CX, [1, 3]], [Rz(5), 3], [CX, [0, 3]], [CX, [1, 3]]],
         [[Rz(1), 3], [CX, [1, 3]], [Rz(3), 3], [CX, [2, 3]], [CX, [0, 3]], [Rz(5), 3],
          [CX, [1, 3]], [Rz(4), 3], [CX, [0, 3]], [Rz(2), 3], [CX, [2, 3]], [CX, [2, 1]]]
         ],
        [4, 1, [2, 1, 0, 3],
         [[CX, [2, 1]], [CX, [0, 3]], [CX, [1, 3]], [Rz(1), 3], [CX, [1, 3]], [CX, [2, 3]],
          [Rz(2), 3], [CX, [0, 3]], [Rz(3), 3], [CX, [1, 3]], [Rz(4), 3]],
         [[CX, [2, 3]], [Rz(3), 3], [CX, [0, 3]], [Rz(2), 3], [CX, [1, 3]], [Rz(1), 3],
          [CX, [0, 3]], [CX, [2, 3]], [Rz(4), 3], [CX, [2, 1]]]
         ],
        [4, 2, [1, 2, 3, 0],
         [[CX, [2, 1]], [CX, [1, 3]], [CX, [2, 3]], [Rz(1), 3], [CX, [1, 3]], [Rz(2), 3], [CX, [0, 3]],
          [Rz(3), 3], [CX, [2, 3]], [CX, [1, 3]], [Rz(4), 3], [CX, [0, 3]], [CX, [1, 3]]],
         [[CX, [2, 3]], [Rz(2), 3], [CX, [0, 3]], [Rz(3), 3], [CX, [1, 3]], [Rz(4), 3],
          [CX, [0, 3]], [CX, [2, 3]], [Rz(1), 3], [CX, [1, 3]], [CX, [2, 1]]]
         ],
        [4, 2, [1, 2, 3, 0],
         [[CX, [2, 1]], [CX, [2, 3]], [CX, [1, 3]], [Rz(1), 3], [CX, [1, 3]], [Rz(2), 3],
          [CX, [0, 3]], [Rz(3), 3], [CX, [2, 3]], [CX, [1, 3]], [Rz(4), 3], [CX, [0, 3]], [CX, [1, 3]]],
         [[CX, [2, 3]], [Rz(2), 3], [CX, [0, 3]], [Rz(3), 3], [CX, [1, 3]], [Rz(4), 3],
          [CX, [0, 3]], [CX, [2, 3]], [Rz(1), 3], [CX, [1, 3]], [CX, [2, 1]]]
         ],
    ]

    ret = []
    for n_qubit, weight, order, tpl, rpl in tpl_list:
        tpl_circ = get_circuit_from_list(n_qubit, tpl)
        rpl_circ = get_circuit_from_list(n_qubit, rpl)
        ret.append(ParameterizedTemplate(DAG(tpl_circ), DAG(rpl_circ), weight=weight, param_order=order))
    return ret



from .unitary_decomposition import UnitaryDecomposition
from .controlled_unitary import ControlledUnitaryDecomposition
from .cartan_kak_decomposition import CartanKAKDecomposition
from .cartan_kak_diagonal_decomposition import CartanKAKDiagonalDecomposition



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch_geometric.nn as gnn


class GnnBlock(nn.Module):
    def __init__(
        self,
        qubit_num: int,
        max_gate_num: int,
        feat_dim: int,
        pool: bool,
    ) -> None:
        super().__init__()

        self._max_qubit_num = qubit_num
        self._max_gate_num = max_gate_num
        self._feat_dim = feat_dim
        self._pool = pool

        self._gc_grp = nn.ModuleList(
            [
                gnn.SAGEConv(
                    in_channels=feat_dim,
                    out_channels=feat_dim,
                    normalize=False,
                ),
                gnn.SAGEConv(
                    in_channels=feat_dim,
                    out_channels=feat_dim,
                    normalize=False,
                ),
            ]
        )

        if pool:
            # TopKPooling is returning (x, edge_index, edge_attr, batch, perm, score[perm])
            self._pool = gnn.TopKPooling(in_channels=feat_dim)

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        batch: Optional[torch.Tensor],
    ):
        for conv in self._gc_grp:
            x = F.leaky_relu(conv(x, edge_index)) + x
        if self._pool:
            x, edge_index, _, batch, _, _ = self._pool(x, edge_index, None, batch, None)
        return x, edge_index, batch


class CircuitGnn(nn.Module):
    def __init__(
        self,
        qubit_num: int,
        max_gate_num: int,
        feat_dim: int,
    ) -> None:
        super().__init__()

        self._max_qubit_num = qubit_num
        self._max_gate_num = max_gate_num
        self._feat_dim = feat_dim

        self._gc_grp = nn.ModuleList(
            [
                GnnBlock(
                    qubit_num=qubit_num,
                    max_gate_num=max_gate_num,
                    feat_dim=feat_dim,
                    pool=False,
                ),
                GnnBlock(
                    qubit_num=qubit_num,
                    max_gate_num=max_gate_num,
                    feat_dim=feat_dim,
                    pool=False,
                ),
                GnnBlock(
                    qubit_num=qubit_num,
                    max_gate_num=max_gate_num,
                    feat_dim=feat_dim,
                    pool=False,
                ),
                GnnBlock(
                    qubit_num=qubit_num,
                    max_gate_num=max_gate_num,
                    feat_dim=feat_dim,
                    pool=False,
                ),
                GnnBlock(
                    qubit_num=qubit_num,
                    max_gate_num=max_gate_num,
                    feat_dim=feat_dim,
                    pool=False,
                ),
            ]
        )

        self._aggr = gnn.aggr.SoftmaxAggregation(learn=False)

    def forward(
        self,
        x: torch.Tensor,
        edge_index: torch.Tensor,
        batch: Optional[torch.Tensor] = None,
    ):
        f = self._feat_dim

        for conv_block in self._gc_grp:
            x, edge_index, batch = conv_block(x, edge_index, batch)

        x = self._aggr(x, batch)  # [b, f]
        x = x.view(-1, f)  # [b, f]
        return x


class GnnMapping(nn.Module):
    def __init__(
        self,
        qubit_num: int,
        max_gate_num: int,
        feat_dim: int,
        action_num: int,
        device: str,
    ) -> None:
        super().__init__()

        self._max_qubit_num = qubit_num
        self._max_gate_num = max_gate_num
        self._feat_dim = feat_dim
        self._action_num = action_num
        self._device = device

        # All gate nodes and virtual node feature embedding.
        self._x_trans = nn.Embedding(
            num_embeddings=qubit_num + 1,
            embedding_dim=feat_dim,
            padding_idx=0,
        )

        self._circ_gnn = CircuitGnn(
            qubit_num=qubit_num,
            max_gate_num=max_gate_num,
            feat_dim=feat_dim * 2,
        )

        f_start = feat_dim * 2
        self._mlp_1 = nn.Sequential(
            nn.Linear(f_start, f_start),
            nn.LeakyReLU(),
            nn.Linear(f_start, f_start // 2),
            nn.LeakyReLU(),
            nn.Linear(f_start // 2, self._action_num),
        )

    def forward(self, x: torch.Tensor, edge_index: torch.Tensor, batch: torch.Tensor):
        f = self._feat_dim
        a = self._action_num

        circ_x = self._x_trans(x).view(-1, 2 * f)
        circ_feat = self._circ_gnn(circ_x, edge_index, batch)  # [b, f]

        x = self._mlp_1(circ_feat).view(-1, a)  # [b, a]
        return x


NnMapping = GnnMapping



"""
Decomposition of SU(4) with Cartan KAK Decomposition
"""

import numpy as np

from QuICT.core.gate import CompositeGate, CX, Ry, Rz, Unitary

# Magic basis
B = (1.0 / np.sqrt(2)) * np.array([[1, 1j, 0, 0],
                                   [0, 0, 1j, 1],
                                   [0, 0, 1j, -1],
                                   [1, -1j, 0, 0]], dtype=complex)


class CartanKAKDecomposition(object):
    """Cartan KAK Decomposition in SU(4)

    ∀ U in SU(4), ∃ KL0, KL1, KR0, KR1 in SU(2), a, b, c in R, s.t.
    U = (KL0⊗KL1).exp(i(a XX + b YY + c ZZ)).(KR0⊗KR1)

    Proof of this proposition in general cases is too 'mathematical' even for TCS
    researchers. [2] gives a proof for U(4), which is a little more friendly for
    researchers without mathematical background. However, be aware that [2] uses
    a different notation.

    The process of this part is taken from [1], while some notation is from [3],
    which is useful in the build_gate.

    Reference:
        [1] https://arxiv.org/abs/0806.4015
        [2] https://arxiv.org/abs/quant-ph/0507171
        [3] https://arxiv.org/abs/quant-ph/0308006
    """

    def __init__(self, eps=1e-15):
        """
        Args:
            eps(float, optional): Eps of decomposition process
        """
        self.eps = eps

    @staticmethod
    def diagonalize_unitary_symmetric(matrix):
        """
        Diagonalize unitary symmetric matrix with real orthogonal matrix

        Args:
            matrix(np.array): unitary symmetric matrix to be diagonalized

        Returns:
            Tuple[np.array, np.array]: Eigenvalues and eigenvectors
        """
        M2 = matrix.copy()
        # D, P = la.eig(M2)  # this can fail for certain kinds of degeneracy
        for i in range(100):  # FIXME: this randomized algorithm is horrendous
            state = np.random.default_rng(i)
            M2real = state.normal() * M2.real + state.normal() * M2.imag
            _, P = np.linalg.eigh(M2real)
            D = P.T.dot(M2).dot(P).diagonal()
            if np.allclose(P.dot(np.diag(D)).dot(P.T), M2, rtol=1.0e-6, atol=1.0e-6):
                break
        else:
            raise ValueError("CartanKAKDecomposition: failed to diagonalize M2")
        return D, P

    @staticmethod
    def tensor_decompose(matrix):
        """
        Decompose U in SU(2)⊗SU(2) to U0, U1 in SU(2), s.t. U = U0⊗U1

        Args:
            matrix(np.array): Matrix to be decomposed

        Returns:
            Tuple(np.array): Decomposed result U0, U1
        """
        U = matrix.copy()
        # Decompose U1
        U1 = U[:2, :2].copy()
        # There is chance that U0[0, 0] == 0 (or more explicitly, is close to 0)
        if np.abs(np.linalg.det(U1)) < 0.1:
            U1 = U[2:, :2].copy()
        # If U0[0, 0], U0[1, 0] are both close to 0, U0 would not be unitary.
        if np.abs(np.linalg.det(U1)) < 0.1:
            raise ValueError("tensor_decompose: U1 failed")
        U1 /= np.sqrt(np.linalg.det(U1))

        # Decompose U0
        U1_inv = np.kron(np.eye(2), U1.T.conj())
        U0_tensor = U1_inv.dot(U)
        U0 = U0_tensor[::2, ::2]
        if np.abs(np.linalg.det(U0)) < 0.9:
            raise ValueError("tensor_decompose: U0 failed")
        U0 /= np.sqrt(np.linalg.det(U0))

        # Final test
        res = np.kron(U0, U1)
        dev = np.abs(np.abs(res.conj(res).T.dot(U).trace()) - 4)
        assert dev < 1e-6, ValueError("tensor_decompose: Final failed")
        return U0, U1

    def execute(self, matrix):
        """
        Decompose a matrix U in SU(4) with Cartan KAK Decomposition to
        a circuit, which contains only 1-qubit gates and CNOT gates.
        The decomposition of Exp(i(a XX + b YY + c ZZ)) may vary a global phase.

        Args:
            matrix(np.array): 4*4 unitary matrix to be decomposed

        Returns:
            CompositeGate: Decomposed gates.

        Reference:
            [1] https://arxiv.org/abs/0806.4015
            [2] https://arxiv.org/abs/quant-ph/0308006
        """
        assert matrix.shape == (4, 4), \
            ValueError("CartanKAKDecomposition: Input must be a 4*4 matrix.")
        assert np.allclose(matrix.T.conj().dot(matrix), np.eye(4), rtol=1e-6, atol=1e-6), \
            ValueError("CartanKAKDecomposition: Input must be a unitary matrix.")

        U = matrix.copy()
        U /= np.linalg.det(U) ** (0.25)
        eps = self.eps

        # Some preparation derived from Cartan involution
        Up = B.T.conj().dot(U).dot(B)
        M2 = Up.T.dot(Up)
        M2.real[abs(M2.real) < eps] = 0.0
        M2.imag[abs(M2.imag) < eps] = 0.0

        # Since M2 is a symmetric unitary matrix, we can diagonalize its real and
        # imaginary part simultaneously. That is, ∃ P in SO(4), s.t. M2 = P.D.P^T,
        # where D is diagonal with unit-magnitude elements.
        D, P = self.diagonalize_unitary_symmetric(M2)

        # Calculated D is usually in U(4) instead of SU(4), therefore d[3] is reset
        # so that D is now in SU(4)
        d = np.angle(D) / 2
        d[3] = -d[0] - d[1] - d[2]
        a = (d[0] + d[2]) / 2
        b = (d[1] + d[2]) / 2
        c = (d[0] + d[1]) / 2

        # Similarly, P could be in O(4) instead of SO(4).
        if np.linalg.det(P) < 0:
            P[:, -1] = -P[:, -1]

        # Now is the time to calculate KL and KR
        KL = B.dot(Up).dot(P).dot(np.diag(np.exp(-1j * d))).dot(B.T.conj())
        KR = B.dot(P.T).dot(B.T.conj())
        KL.real[abs(KL.real) < eps] = 0.0
        KL.imag[abs(KL.imag) < eps] = 0.0
        KR.real[abs(KR.real) < eps] = 0.0
        KR.imag[abs(KR.imag) < eps] = 0.0
        KL0, KL1 = self.tensor_decompose(KL)
        KR0, KR1 = self.tensor_decompose(KR)

        KL0 = KL0.dot(Rz(-np.pi / 2).matrix)
        KR1 = Rz(np.pi / 2).matrix.dot(KR1)
        gates = CompositeGate()
        with gates:
            Unitary(KR0) & 0
            Unitary(KR1) & 1
            CX & [1, 0]
            Rz(np.pi / 2 - 2 * c) & 0
            Ry(np.pi / 2 - 2 * a) & 1
            CX & [0, 1]
            Ry(2 * b - np.pi / 2) & 1
            CX & [1, 0]
            Unitary(KL0) & 0
            Unitary(KL1) & 1

        return gates



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/3/22 10:31
# @Author  : Han Yu
# @File    : _qasmInterface.py

import os
from collections import OrderedDict
from configparser import ConfigParser

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.core.utils import GateType
from QuICT.lib import Qasm
from QuICT.tools.exception.core import QASMError

from .basic_interface import BasicInterface


class qasm_qreg(object):
    """ the target bit in the qasm grammer tree

    """

    def __init__(self, index_list, name):
        self.index = index_list
        self.name = name


class qasm_creg(object):
    """ the control bit in the qasm grammer tree

    """

    def __init__(self, index_list, name):
        self.index = index_list
        self.name = name


class OPENQASMInterface(BasicInterface):
    # qasm mapping to QuICT
    standard_extension = {
        "u1": GateType.u1,
        "u2": GateType.u2,
        "u3": GateType.u3,
        "U": GateType.u3,
        "x": GateType.x,
        "y": GateType.y,
        "z": GateType.z,
        "t": GateType.t,
        "tdg": GateType.tdg,
        "s": GateType.s,
        "sdg": GateType.sdg,
        "sx": GateType.sx,
        "sy": GateType.sy,
        "sw": GateType.sw,
        "p": GateType.phase,
        "phase": GateType.gphase,
        "swap": GateType.swap,
        "iswap": GateType.iswap,
        "iswapdg": GateType.iswapdg,
        "sqiswap": GateType.sqiswap,
        "rx": GateType.rx,
        "ry": GateType.ry,
        "rz": GateType.rz,
        "id": GateType.id,
        "h": GateType.h,
        "hy": GateType.hy,
        "cx": GateType.cx,
        "ccx": GateType.ccx,
        "ccz": GateType.ccz,
        "cy": GateType.cy,
        "cz": GateType.cz,
        "ch": GateType.ch,
        "crz": GateType.crz,
        "ccrz": GateType.ccrz,
        "rzz": GateType.rzz,
        "rxx": GateType.rxx,
        "ryy": GateType.ryy,
        "rzx": GateType.rzx,
        "cu1": GateType.cu1,
        "cu3": GateType.cu3,
        "cswap": GateType.cswap,
        "fsim": GateType.fsim
    }

    token = None

    DEFAULT_QUICT_PATH = os.path.join(os.path.expanduser("~"),
                                      '.QuICT')

    DEFAULT_QUICT_FILE = os.path.join(os.path.expanduser("~"),
                                      '.QuICT', 'accounts.ini')

    @staticmethod
    def load_circuit(circuit: Circuit):
        instance = OPENQASMInterface()
        instance.circuit = circuit
        instance.analyse_code_from_circuit()
        return instance

    @staticmethod
    def load_file(filename: str):
        instance = OPENQASMInterface()
        instance.ast = Qasm(filename).parse()
        instance.analyse_circuit_from_ast(instance.ast)
        return instance

    @staticmethod
    def load_string(qasm: str):
        instance = OPENQASMInterface()
        instance.ast = Qasm(data=qasm).parse()
        instance.analyse_circuit_from_ast(instance.ast)
        return instance

    def __init__(self):
        super().__init__()
        self.circuit = None
        self.ast = None
        self.qasm = None

        self.qbits = 0
        self.cbits = 0
        self.valid_circuit = False
        self.valid_qasm = False
        self.gates = OrderedDict()
        self.node_gates = []
        self.qregs = {}
        self.cregs = {}
        self.arg_stack = [{}]
        self.bit_stack = [{}]
        self.circuit_gates = []
        self.version = None

    def analyse_circuit_from_ast(self, node):
        self.valid_circuit = True
        self.analyse_node(node)
        if self.valid_circuit:
            self.circuit = Circuit(self.qbits)
            self.circuit.extend(self.circuit_gates)

    def analyse_code_from_circuit(self):
        self.valid_qasm = True
        self.qasm = self.circuit.qasm()
        if self.qasm == "error":
            self.qasm = None
            self.valid_qasm = False

    def output_qasm(self, filename=None):
        if not self.valid_qasm or self.qasm is None:
            if self.circuit is None:
                return
            self.qasm = self.circuit.qasm()
            if self.qasm != 'error':
                self.valid_qasm = True
            else:
                return False
        if filename is None:
            print(self.qasm)
        else:
            with open(filename, 'w+') as file:
                file.write(self.qasm)

    def enable_path(self):
        if not os.path.exists(self.DEFAULT_QUICT_PATH):
            os.mkdir(self.DEFAULT_QUICT_PATH)

    def load_token(self):
        self.enable_path()
        config_parser = ConfigParser()
        config_parser.read(self.DEFAULT_QUICT_FILE)
        if 'account' in config_parser:
            dic = dict(config_parser.items('account'))
            if 'token' in dic:
                self.token = dic['token']

    def save_token(self, token):
        self.token = token
        self.enable_path()
        config_parser = ConfigParser()
        config_parser['account'] = {
            'token': token,
        }
        with open(self.DEFAULT_QUICT_FILE, 'w') as f:
            config_parser.write(f)

    def analyse_node(self, node):
        if not self.valid_circuit:
            return
        if node.type == "program":
            for child in node.children:
                self.analyse_node(child)

        elif node.type == "qreg":
            self.qregs[node.name] = qasm_qreg([i for i in range(self.qbits, self.qbits + node.index)], node.name)
            self.qbits += node.index

        elif node.type == "creg":
            self.cregs[node.name] = qasm_creg([i for i in range(self.cbits, self.cbits + node.index)], node.name)
            self.cbits += node.index

        elif node.type == "id_list":
            return [self.get_analyse_id(child) for child in node.children]

        elif node.type == "primary_list":
            return [self.get_analyse_id(child) for child in node.children]

        elif node.type == "gate":
            self.analyse_gate(node)

        elif node.type == "custom_unitary":
            self.analyse_custom(node)

        elif node.type == "universal_unitary":
            QASMError("universal_unitary is deprecated", node.line, node.file)

        elif node.type == "cnot":
            self.analyse_cnot(node)

        elif node.type == "expression_list":
            return node.children

        elif node.type == "measure":
            self.analyse_measure(node)

        elif node.type == "format":
            self.version = node.version()

        elif node.type == "barrier":
            pass

        elif node.type == "reset":
            self.analyse_reset(node)

        elif node.type == "if":
            self.analyse_if(node)

        elif node.type == "opaque":
            self.analyse_opaque(node)

        else:
            QASMError("QASM grammer error", node.line, node.file)

    def analyse_gate(self, node):
        self.gates[node.name] = {}
        de_gate = self.gates[node.name]
        de_gate["n_args"] = node.n_args()
        de_gate["n_bits"] = node.n_bits()
        if node.n_args() > 0:
            de_gate["args"] = [element.name for element in node.arguments.children]
        else:
            de_gate["args"] = []
        de_gate["bits"] = [child.name for child in node.bitlist.children]
        if node.name in self.standard_extension:
            return
        de_gate["body"] = node.body

    def analyse_custom(self, node):
        name = node.name
        if node.arguments is not None:
            args = self.analyse_node(node.arguments)
        else:
            args = []
        bits = [
            self.get_analyse_id(node_element)
            for node_element in node.bitlist.children
        ]
        if name in self.gates:
            gargs = self.gates[name]["args"]
            gbits = self.gates[name]["bits"]

            maxidx = max(map(len, bits))
            for idx in range(maxidx):
                self.arg_stack.append({gargs[j]: args[j]
                                       for j in range(len(gargs))})
                element = [idx * x for x in
                           [len(bits[j]) > 1 for j in range(len(bits))]]
                self.bit_stack.append({gbits[j]: bits[j][element[j]]
                                       for j in range(len(gbits))})
                self.analyse_name(name,
                                  gargs,
                                  gbits)
                self.arg_stack.pop()
                self.bit_stack.pop()
        else:
            raise QASMError("undefined gate:", node.line, node.file)

    def analyse_cnot(self, node):
        id0 = self.get_analyse_id(node.children[0])
        id1 = self.get_analyse_id(node.children[1])
        if not (len(id0) == len(id1) or len(id0) == 1 or len(id1) == 1):
            raise QASMError("the number of bits unmatched:", node.line, node.file)

        maxidx = max([len(id0), len(id1)])
        for idx in range(maxidx):
            if len(id0) > 1 and len(id1) > 1:
                qubit_idxes = [id0[idx], id1[idx]]
            elif len(id0) > 1:
                qubit_idxes = [id0[idx], id1[0]]
            else:
                qubit_idxes = [id0[0], id1[idx]]

            cx_gate = build_gate(GateType.cx, qubit_idxes)
            self.circuit_gates.append(cx_gate)

    def analyse_measure(self, node):
        id0 = self.get_analyse_id(node.children[0])
        id1 = self.get_analyse_id(node.children[1])
        if len(id0) != len(id1):
            raise QASMError("the number of bits of registers unmatched:", node.line, node.file)

        for idx, _ in zip(id0, id1):
            m_gate = build_gate(GateType.measure, [idx])
            self.circuit_gates.append(m_gate)

    def analyse_reset(self, node):
        id0 = self.get_analyse_id(node.children[0])
        for i, _ in enumerate(id0):
            r_gate = build_gate(GateType.reset, [id0[i]])
            self.circuit_gates.append(r_gate)

    def analyse_if(self, node):
        print("if op is not supported:{}", node.type)
        self.valid_circuit = False

    def analyse_opaque(self, node):
        pass

    def analyse_name(self, name, gargs, gbits):
        if name in self.standard_extension:
            pargs = [self.arg_stack[-1][s].sym(self.arg_stack[:-1]) for s in gargs]
            targs = [self.bit_stack[-1][s] for s in gbits]
            type = self.standard_extension[name]
            gate = build_gate(type, targs, pargs)
            self.circuit_gates.append(gate)
        else:
            body = self.gates[name]['body']
            for child in body.children:
                if child.arguments is not None:
                    args = self.analyse_node(child.arguments)
                else:
                    args = []
                bits = [
                    self.get_analyse_id(node_element)
                    for node_element in child.bitlist.children
                ]
                gargs = self.gates[child.name]["args"]
                gbits = self.gates[child.name]["bits"]

                maxidx = max(map(len, bits))
                for idx in range(maxidx):
                    self.arg_stack.append({gargs[j]: args[j]
                                           for j in range(len(gargs))})
                    element = [idx * x for x in
                               [len(bits[j]) > 1 for j in range(len(bits))]]
                    self.bit_stack.append({gbits[j]: bits[j][element[j]]
                                           for j in range(len(gbits))})
                    self.analyse_name(child.name,
                                      gargs,
                                      gbits)
                    self.arg_stack.pop()
                    self.bit_stack.pop()

    def get_analyse_id(self, node):
        if node.name in self.qregs:
            reg = self.qregs[node.name]
        elif node.name in self.cregs:
            reg = self.cregs[node.name]
        elif self.bit_stack[-1] and node.name in self.bit_stack[-1]:
            reg = self.bit_stack[-1][node.name]
        else:
            raise QASMError("expected qreg or creg name:", node.line, node.file)

        if node.type == "indexed_id":
            return [reg.index[node.index]]
        elif node.type == "id":
            if not self.bit_stack[-1]:
                return [bit for bit in reg.index]
            else:
                if node.name in self.bit_stack[-1]:
                    return [self.bit_stack[-1][node.name]]
                raise QASMError("expected qreg or creg name:", node.line, node.file)
        return None



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/1/11 2:04 下午
# @Author  : Han Yu
# @File    : __init__.py

from .grover import Grover
from .search_with_prior_knowledge import GroverWithPriorKnowledge
from .partial_grover import PartialGrover



import os
import numpy as np


class Result:
    """ Data structure class for simulator result

    Args:
        mode (str): The simulator mode, usually be {device-backend}.
        shots (int): The running times; must be a positive integer.
        options (dict): Optional parameters for the simulator.
    """
    def __init__(
        self,
        device: str,
        backend: str,
        precision: str,
        circuit_record: bool,
        amplitude_record: bool,
        options: dict,
        output_path: str = None
    ):
        self.device = device
        self.backend = backend
        self.precision = precision
        self._circuit_record = circuit_record
        self._amplitude_record = amplitude_record
        self.options = options

        self.counts = {}
        self.state_vector = None
        self.density_matrix = None

        # prepare output path
        self._dump_folder = True if self._amplitude_record or self._circuit_record else False
        if self._dump_folder:
            self._output_path = self._prepare_output_file(output_path)

    def __str__(self):
        return f"Device: {self.device}\nBackend: {self.backend}\n" + \
            f"Options: {self.options}\nResults: {self.counts}"

    def __dict__(self):
        return {
            "id": self.id,
            "shots": self.shots,
            "backend_info": {
                "device": self.device,
                "backend": self.backend,
                "options": self.options
            },
            "data": {
                "counts": self.counts,
                "state_vector": self.state_vector,
                "density_matrix": self.density_matrix
            }
        }

    def _prepare_output_file(self, output_path):
        """ Prepare output path. """
        if output_path is None:
            curr_path = os.getcwd()
            output_path = os.path.join(curr_path, "output")

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        return output_path

    def record_sample(self, result: list):
        """ Record circuit's result

        Args:
            result (list): The sample of measured result from given circuit.
        """
        self.shots = sum(result)
        for i in range(len(result)):
            bit_idx = "{0:0b}".format(i)
            bit_idx = bit_idx.zfill(int(np.log2(len(result))))
            self.counts[bit_idx] = result[i]

        if self._dump_folder:
            with open(f"{self._output_path}/result_{self.id}.log", "w") as of:
                of.write(str(self.__dict__()))

    def record_circuit(self, circuit):
        """ dump the circuit. """
        self.id = circuit.name
        if self._circuit_record:
            with open(f"{self._output_path}/circuit_{self.id}.qasm", "w") as of:
                of.write(circuit.qasm())

    def record_amplitude(self, amplitude):
        """ dump the circuit. """
        if self.device == "GPU":
            amplitude = amplitude.get()

        if self._amplitude_record:
            np.savetxt(f"{self._output_path}/amp_{self.id}.txt", amplitude)

        if self.backend == "density_matrix":
            self.density_matrix = amplitude.copy()
        else:
            self.state_vector = amplitude.copy()



from enum import Enum

from ._operator import Operator


class DataSwitchType(Enum):
    all = "ALL"
    half = "HALF"
    ctarg = "CTARGS"


class DataSwitch(Operator):
    def __init__(self, destination: int, type: DataSwitchType, switch_condition: dict = None):
        super().__init__(targets=1)
        self._destination = destination
        self._type = type
        self._cond = switch_condition

    def qasm(self):
        return f"{self.destination}+{self.type}+{self._cond}"

    @property
    def destination(self):
        return self._destination

    @property
    def type(self):
        return self._type

    @property
    def switch_condition(self):
        return self._cond



# qcompute package
from QCompute import *
from QCompute.QPlatform.QOperation import FixedGate
# qiskit package
from qiskit import IBMQ
from qiskit import QuantumCircuit
from qiskit import transpile

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.simulation.remote import RemoteSimulator


class QuantumLeafSimulator(RemoteSimulator):
    """ The QCompute simulator, used to connect the remote simulator by QCompute.

    Args:
        token ([str]): The token to connect the QCompute simulator.
        backend ([str]): The backend for the remote simulator.
        shots ([int]): The running times; must be a positive integer.
    """

    __BACKEND = [   # simulators
        'cloud_aer_at_bd',
        'cloud_baidu_sim2_earth',
        'cloud_baidu_sim2_heaven',
        'cloud_baidu_sim2_lake',
        'cloud_baidu_sim2_thunder',
        'cloud_baidu_sim2_water',
        'cloud_baidu_sim2_wind',
        'cloud_iopcas'  # QPU
    ]

    __GATES = [
        FixedGate.ID,
        FixedGate.X,
        FixedGate.Y,
        FixedGate.Z,
        FixedGate.H,
        FixedGate.S,
        FixedGate.SDG,
        FixedGate.T,
        FixedGate.TDG,
        FixedGate.CX,
        FixedGate.CY,
        FixedGate.CZ,
        FixedGate.CH,
        FixedGate.SWAP,
        FixedGate.CCX,
        FixedGate.CSWAP
    ]

    def __init__(self, token: str, backend: str = 'cloud_aer_at_bd', shots: int = 1):
        RemoteSimulator.__init__(self, backend, shots)
        if backend not in QuantumLeafSimulator.__BACKEND:
            raise ValueError(
                f"Unsupportted backend in QuantumLeaf Simulator. Please use one of {self.__BACKEND}."
            )

        # Create qcompute backend
        Define.hubToken = token
        self._env = QEnv()

    def _initial_circuit(self, circuit: Circuit, use_previous: bool = False):
        """ Initial QCompute's quantum circuit from QuICT circuit.

        Args:
            circuit ([Circuit]): The QuICT circuit.
            use_previous ([bool]): Using previous vector state or not, default to False.

        Raises:
            ValueError: Unsupportted quantum gate

        Returns:
            [object]: The QCompute simulator.
        """
        if use_previous:
            qubits = self._env.Q
        else:
            qubits = self._env.Q.createList(circuit.width())

        gates = circuit.gates
        for gate in gates:
            if isinstance(gate, IDGate):
                targ = gate.targ
                FixedGate.ID(qubits[targ])
            elif isinstance(gate, XGate):
                targ = gate.targ
                FixedGate.X(qubits[targ])
            elif isinstance(gate, YGate):
                targ = gate.targ
                FixedGate.Y(qubits[targ])
            elif isinstance(gate, ZGate):
                targ = gate.targ
                FixedGate.Z(qubits[targ])
            elif isinstance(gate, HGate):
                targ = gate.targ
                FixedGate.H(qubits[targ])
            elif isinstance(gate, SGate):
                targ = gate.targ
                FixedGate.S(qubits[targ])
            elif isinstance(gate, SDaggerGate):
                targ = gate.targ
                FixedGate.SDG(qubits[targ])
            elif isinstance(gate, TGate):
                targ = gate.targ
                FixedGate.T(qubits[targ])
            elif isinstance(gate, TDaggerGate):
                targ = gate.targ
                FixedGate.TDG(qubits[targ])
            elif isinstance(gate, CXGate):
                carg = gate.carg
                targ = gate.targ
                FixedGate.CX(qubits[carg], qubits[targ])
            elif isinstance(gate, CYGate):
                carg = gate.carg
                targ = gate.targ
                FixedGate.CY(qubits[carg], qubits[targ])
            elif isinstance(gate, CZGate):
                carg = gate.carg
                targ = gate.targ
                FixedGate.CZ(qubits[carg], qubits[targ])
            elif isinstance(gate, CHGate):
                carg = gate.carg
                targ = gate.targ
                FixedGate.CH(qubits[carg], qubits[targ])
            elif isinstance(gate, SwapGate):
                targs = gate.targs
                FixedGate.SWAP(qubits[targs[0]], qubits[targs[1]])
            elif isinstance(gate, CCXGate):
                cargs = gate.cargs
                targ = gate.targ
                FixedGate.CCX(qubits[cargs[0]], qubits[cargs[1]], qubits[targ])
            elif isinstance(gate, CSwapGate):
                carg = gate.carg
                targs = gate.targs
                FixedGate.CSWAP(qubits[carg], qubits[targs[0]], qubits[targs[1]])
            elif isinstance(gate, MeasureGate):
                targ = gate.targ
                MeasureZ([qubits[targ]], [targ])
            else:
                raise ValueError("Unsupportted quantum gate in QuantumLeaf Simulator.")

        return self._env

    def run(self, circuit, use_previous):
        """ start simulator with given circuit

        Args:
            circuit (Circuit): The quantum circuits.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.

        return:
            [data]: The experience result.
        """
        self._env.backend(self._backend)
        env = self._initial_circuit(circuit, use_previous)
        result = env.commit(self._shots)

        return result

    def get_gates(self):
        """ The supportted quantum gates in QCompute.

        Returns:
            [dict]: quantum gates with matrix.
        """
        gates = {}
        for gate in QuantumLeafSimulator.__GATES:
            gates[gate.name] = gate.getMatrix()

        return gates


class QiskitSimulator(RemoteSimulator):
    """ The Qiskit simulator, used to connect the remote simulator by Qiskit.

    Args:
        token ([str]): The token to connect the Qiskit simulator.
        backend ([str]): The backend for the remote simulator.
        shots ([int]): The running times; must be a positive integer.
    """

    __BACKEND = [
        'ibmq_qasm_simulator',
        'ibmq_kolkata',
        'ibmq_mumbai',
        'ibmq_dublin',
        'ibmq_hanoi',
        'ibmq_cairo',
        'ibmq_manhattan',
        'ibmq_brooklyn',
        'ibmq_toronto',
        'ibmq_sydney',
        'ibmq_guadalupe',
        'ibmq_casablanca',
        'ibmq_lagos',
        'ibmq_nairobi',
        'ibmq_santiago',
        'ibmq_manila',
        'ibmq_bogota',
        'ibmq_jakarta',
        'ibmq_quito',
        'ibmq_belem',
        'ibmq_lima',
        'ibmq_armonk',
    ]

    def __init__(self, token, backend: str = 'ibmq_qasm_simulator', shots: int = 1):
        RemoteSimulator.__init__(self, backend, shots)
        if backend not in QiskitSimulator.__BACKEND:
            raise ValueError(
                f"Unsupportted backend in Qiskit Simulator. Please use one of {self.__BACKEND}."
            )

        # create qiskit backend
        self._provider = IBMQ.enable_account(token)
        self._backend = self._provider.get_backend(backend)

    def _initial_circuit(self, circuit, use_previous: bool = False):
        """ Initial Qiskit's quantum circuit from QuICT circuit.

        Args:
            circuit ([Circuit]): The QuICT circuit.
            use_previous ([bool]): Using previous vector state or not, default to False.

        Returns:
            [object]: The Qiskit circuit.
        """
        qasm_str = circuit.qasm()
        translated_circuit = QuantumCircuit.from_qasm_str(qasm_str)
        compiled_circuit = transpile(translated_circuit, self._backend)

        return compiled_circuit

    def run(self, circuit, use_previous):
        """ start simulator with given circuit

        Args:
            circuit (Circuit): The quantum circuits.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.

        return:
            [data]: The experience result.
        """
        compiled_circuit = self._initial_circuit(circuit, use_previous)
        result = self._backend.run(compiled_circuit, shots=self._shots)

        return result.result().data()



"""
Another version of decomposition of SU(4) with Cartan KAK Decomposition,
which is specially designed for the optimization of unitary transform.
"""

import numpy as np

from QuICT.core.gate import CompositeGate, CX, Rx, Rz, Unitary
from .cartan_kak_decomposition import CartanKAKDecomposition

# Magic basis
B = (1.0 / np.sqrt(2)) * np.array([[1, 1j, 0, 0],
                                   [0, 0, 1j, 1],
                                   [0, 0, 1j, -1],
                                   [1, -1j, 0, 0]], dtype=complex)


class CartanKAKDiagonalDecomposition(object):
    def __init__(self, eps=1e-15):
        """
        Args:
            eps(float, optional): Eps of decomposition process
        """
        self.eps = eps

    def execute(self, matrix):
        """
        Decompose a matrix U in SU(4) with Cartan KAK Decomposition. Unlike the
        original version, now the result circuit has a two-qubit gate whose
        matrix is diagonal at the edge, which is useful in the optimization.
        The process is taken from [2] Proposition V.2 and Theorem VI.3,
        while the Cartan KAK Decomposition process is refined from [1].

        Args:
            matrix(np.array): 4*4 unitary matrix to be decomposed

        Returns:
            CompositeGate: Decomposed gates.

        Reference:
            [1] https://arxiv.org/abs/0806.4015
            [2] https://arxiv.org/abs/quant-ph/0308033
        """
        sy2 = np.array([[0, 0, 0, -1],
                        [0, 0, 1, 0],
                        [0, 1, 0, 0],
                        [-1, 0, 0, 0]], dtype=complex)

        # Proposition V.2
        U = matrix.copy()
        U /= np.linalg.det(U) ** 0.25
        gUTT = U.T.dot(sy2).dot(U).dot(sy2).T
        denominator = (gUTT[0, 0] - gUTT[1, 1] - gUTT[2, 2] + gUTT[3, 3]).real
        if np.isclose(denominator, 0):
            psi = np.pi / 2
        else:
            numerator = (gUTT[0, 0] + gUTT[1, 1] + gUTT[2, 2] + gUTT[3, 3]).imag
            psi = np.arctan(numerator / denominator)

        gates_Delta = CompositeGate()
        with gates_Delta:
            CX & [0, 1]
            Rz(psi) & 1
            CX & [0, 1]
        Delta = gates_Delta.matrix()
        U = U.dot(Delta)

        # Refined Cartan KAK Decomposition for U (because we have known Ud here!)
        # Some preparation derived from Cartan involution
        Up = B.T.conj().dot(U).dot(B)
        M2 = Up.T.dot(Up)
        M2.real[abs(M2.real) < self.eps] = 0.0
        M2.imag[abs(M2.imag) < self.eps] = 0.0

        # Since M2 is a symmetric unitary matrix, we can diagonalize its real and
        # imaginary part simultaneously. That is, ∃ P in SO(4), s.t. M2 = P.D.P^T,
        # where D is diagonal with unit-magnitude elements.
        D, P = CartanKAKDecomposition.diagonalize_unitary_symmetric(M2)
        d = np.angle(D) / 2

        # A special case occurs when there are pairs of -1 in D
        if np.any(np.isclose(D, -1)):
            if np.count_nonzero(np.isclose(D, -1)) == 2:
                d[np.where(np.isclose(D, -1))] = np.pi / 2, -np.pi / 2
            elif np.count_nonzero(np.isclose(D, -1)) == 4:
                d[:] = np.pi / 2, np.pi / 2, -np.pi / 2, -np.pi / 2
            else:
                raise ValueError('Odd number of -1 in D')

        # Refinement time, by some mathematics we know that d here must be a rearragement
        # of d_Ud. However, the diagonalization process does not guarantee that they are
        # correctly sorted.
        order = np.argsort(d)[::-1]
        d[:] = d[order]
        P[:, :] = P[:, order]
        a = (d[0] + d[2]) / 2
        b = (d[1] + d[2]) / 2
        c = (d[0] + d[1]) / 2
        assert np.isclose(b, 0)

        # P could be in O(4) instead of SO(4).
        if np.linalg.det(P) < 0:
            P[:, -1] = -P[:, -1]

        # Now is the time to calculate KL and KR
        KL = B.dot(Up).dot(P).dot(np.diag(np.exp(-1j * d))).dot(B.T.conj())
        KR = B.dot(P.T).dot(B.T.conj())
        KL.real[abs(KL.real) < self.eps] = 0.0
        KL.imag[abs(KL.imag) < self.eps] = 0.0
        KR.real[abs(KR.real) < self.eps] = 0.0
        KR.imag[abs(KR.imag) < self.eps] = 0.0
        KL0, KL1 = CartanKAKDecomposition.tensor_decompose(KL)
        KR0, KR1 = CartanKAKDecomposition.tensor_decompose(KR)

        # Finally we could combine everything together
        gates = CompositeGate()
        with gates:
            Unitary(Delta.conj()) & [0, 1]
            Unitary(KR0) & 0
            Unitary(KR1) & 1
            CX & [0, 1]
            Rx(-2 * a) & 0
            Rz(-2 * c) & 1
            CX & [0, 1]
            Unitary(KL0) & 0
            Unitary(KL1) & 1

        return gates



"""
Optimize Clifford circuits with template matching and symbolic peephole optimization
"""

import itertools
import random

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import (CX, CY, CZ, CompositeGate, GateType, H, S,
                             S_dagger, X, Z)
from QuICT.qcda.optimization.commutative_optimization import \
    CommutativeOptimization
from QuICT.qcda.synthesis.gate_transform.transform_rule import (cy2cx_rule,
                                                                cz2cx_rule)
from QuICT.qcda.utility import OutputAligner, PauliOperator


class SymbolicCliffordOptimization(object):
    """
    Implement the Clifford circuit optimization process described in Reference, which
    consists of the following 4 steps:
    1. Partition the circuit into compute and Pauli stages
    2. Apply template matching to the compute stage
    3. Apply symbolic peephole optimization to the compute stage
    4. Optimize the 1-qubit gate count

    Reference:
        https://arxiv.org/abs/2105.02291
    """
    def __init__(self, control_sets=None):
        """
        Args:
            control_sets(list, optional): containing several control_set, which is list of qubit,
                CX coupling control_set and the complement would be decoupled
        """
        assert control_sets is None or isinstance(control_sets, list),\
            TypeError('control_set must be list of qubit')
        self.control_sets = control_sets

    def __repr__(self):
        return 'SymbolicCliffordOptimization()'

    @OutputAligner()
    def execute(self, gates: CompositeGate):
        """
        Args:
            gates(Circuit/CompositeGate): the Clifford Circuit/CompositeGate to be optimized

        Returns:
            CompositeGate: the Clifford CompositeGate after optimization
        """
        width = gates.width()
        if isinstance(gates, Circuit):
            gates = CompositeGate(gates=gates.gates)
        assert isinstance(gates, CompositeGate),\
            TypeError('Invalid input(Circuit/CompositeGate)')
        for gate in gates:
            # FIXME log output here
            if not gate.is_clifford():
                return gates

        if self.control_sets is None:
            self.control_sets = list(itertools.combinations(range(width), 2))

        compute, global_pauli = self.partition(gates, width)

        size = compute.size()
        while True:
            control_set = list(random.choice(self.control_sets))
            compute, global_pauli = self.circular_optimization(compute, width, global_pauli)
            compute = self.cx_reverse(compute, control_set)
            compute, global_pauli = self.circular_optimization(compute, width, global_pauli)
            compute = self.reorder(compute)
            compute = self.symbolic_peephole_optimization(compute, control_set)
            compute, global_pauli = self.circular_optimization(compute, width, global_pauli)
            if compute.size() == size:
                break
            size = compute.size()

        compute.extend(global_pauli.gates(keep_phase=True))
        return compute

    def reorder(self, gates: CompositeGate):
        """
        For CompositeGate with CX, H, S only, reorder gates so that S appears as late as possible
        """
        for gate in gates:
            assert gate.type in [GateType.cx, GateType.h, GateType.s],\
                ValueError('Only CX, H, S gates are allowed in reorder')
        width = gates.width()
        gates_reorder = CompositeGate()
        S_stack = [[] for _ in range(width)]
        for gate in gates:
            if gate.type == GateType.s:
                S_stack[gate.targ].append(gate)
            else:
                gates_reorder.extend(S_stack[gate.targ])
                gates_reorder.append(gate)
                S_stack[gate.targ] = []
        for qubit in range(width):
            gates_reorder.extend(S_stack[qubit])
        return gates_reorder

    def cx_reverse(self, gates: CompositeGate, control_set: list):
        """
        For CX in the CompositeGate, if its target qubit is in the control_set while its control qubit
        is not, the control and target qubit will be reversed by adding H gates.
        """
        gates_reverse = CompositeGate()
        for gate in gates:
            if gate.type == GateType.cx and gate.carg not in control_set and gate.targ in control_set:
                with gates_reverse:
                    H & gate.carg
                    H & gate.targ
                    CX & [gate.targ, gate.carg]
                    H & gate.carg
                    H & gate.targ
            else:
                gates_reverse.append(gate)
        return gates_reverse

    def circular_optimization(self, gates: CompositeGate, width: int, global_pauli: PauliOperator):
        # TODO: More optimization to be added here
        CO = CommutativeOptimization(deparameterization=True)
        while True:
            gates_opt = CO.execute(gates)
            compute, pauli = self.partition(gates_opt, width)
            global_pauli = pauli.combine(global_pauli)
            if compute.size() == gates.size() and gates_opt.size() == gates.size():
                break
            gates = compute
        return compute, global_pauli

    @staticmethod
    def partition(gates: CompositeGate, width=None):
        """
        Partition the CompositeGate into compute and Pauli stages, where compute stage contains
        only CX, H, S, Sdg gates and Pauli stage contains only Pauli gates.

        Args:
            gates(CompositeGate): Clifford CompositeGate
            width(int, optional): width of the CompositeGate

        Returns:
            CompositeGate, PauliOperator: compute stage and Pauli stage
        """
        for gate in gates:
            assert gate.is_clifford(), ValueError('Only Clifford CompositeGate')
        if width is None:
            width = gates.width()
        pauli = PauliOperator([GateType.id for _ in range(width)])

        compute = CompositeGate()
        for gate in gates:
            if gate.is_pauli():
                pauli.combine_one_gate(gate.type, gate.targ)
                continue
            if gate.type == GateType.cx or gate.type == GateType.h or gate.type == GateType.s:
                pauli.conjugate_act(gate)
                compute.append(gate.copy())
                continue
            if gate.type == GateType.sdg:
                pauli.combine_one_gate(GateType.z, gate.targ)
                pauli.conjugate_act(gate.inverse())
                compute.append(gate.inverse())
                continue

        return compute, pauli

    @classmethod
    def symbolic_peephole_optimization(cls, gates: CompositeGate, control_set: list):
        """
        Symbolic Pauli gate gives another expression for controlled Pauli gates.
        By definition, a controlled-U gate CU means:
            if the control qubit is |0>, do nothing;
            if the control qubit is |1>, apply U to the target qubit.
        In general, CU = ∑_v |v><v| ⊗ U^v, where U^v is called a symbolic gate.

        Here we focus only on symbolic Pauli gates and symbolic phase.
        By decoupling CNOT gates with projectors and symbolic Pauli gates, optimization
        rules of 1-qubit gates could be used to optimize Clifford circuits.

        Args:
            gates(CompositeGate): CompositeGate with CX, H, S gates only where the CX coupling control_set
                and the complement must have the control qubit in control_set
            control_set(list): list of qubit, CX coupling control_set and the complement would be decoupled

        Returns:
            CompositeGate: CompositeGate after optimization
        """
        for gate in gates:
            assert gate.type in [GateType.cx, GateType.h, GateType.s],\
                ValueError('Only CX, H, S gates are allowed in this optimization')
            if gate.type == GateType.cx:
                assert not (gate.carg not in control_set and gate.targ in control_set),\
                    ValueError('Coupling CX must have the control qubit in control_set')

        gates_opt = CompositeGate()
        current = CompositeGate()
        control = None
        for gate in gates:
            if current.size() != 0:
                if ((gate.type != GateType.cx and gate.targ != control) or
                   (gate.type == GateType.cx and (gate.carg == control or gate.carg not in control_set))):
                    current.append(gate)
                else:
                    gates_opt.extend(cls.local_symbolic_optimization(current, control))
                    current = CompositeGate()
            if current.size() == 0:
                if (gate.type != GateType.cx or
                   (gate.type == GateType.cx and gate.carg not in control_set)):
                    gates_opt.append(gate)
                else:
                    current.append(gate)
                    control = gate.carg

        if current.size() != 0:
            gates_opt.extend(cls.local_symbolic_optimization(current, control))
        return gates_opt

    @classmethod
    def local_symbolic_optimization(cls, gates: CompositeGate, control: int):
        """
        Inner method for symbolic peephole optimization
        """
        for gate in gates:
            assert gate.targ != control, ValueError('control qubit should not be targeted')

        symbolic_gates = CompositeGate()
        for gate in gates:
            # symbolize coupling cx
            if gate.type == GateType.cx and gate.carg == control:
                symbolic_gates.append(X & gate.targ)
            else:
                symbolic_gates.append(gate)

        # pauli here is symbolic pauli, including symbolic phase
        compute, pauli = cls.partition(symbolic_gates)

        # such partition may cause negative optimization, if so, return the original gates
        if gates.count_2qubit_gate() < compute.count_2qubit_gate() + pauli.hamming_weight:
            return gates

        # restore the symbolic pauli
        for qubit in range(pauli.width):
            if qubit == control:
                assert pauli.operator[qubit] == GateType.id
                continue
            if pauli.operator[qubit] == GateType.x:
                compute.append(CX & [control, qubit])
            if pauli.operator[qubit] == GateType.y:
                compute.extend(cy2cx_rule(CY & [control, qubit]))
            if pauli.operator[qubit] == GateType.z:
                compute.extend(cz2cx_rule(CZ & [control, qubit]))
        # restore the symbolic phase
        if np.isclose(pauli.phase, 1j):
            compute.append(S & control)
        if np.isclose(pauli.phase, -1):
            compute.append(Z & control)
        if np.isclose(pauli.phase, -1j):
            compute.append(S_dagger & control)

        return compute



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/5 10:55
# @Author  : Zhu Qinlin
# @File    : tmvh.py

from QuICT.core.gate import CX, CCX, CompositeGate, X


def peres_gate(gate_set, a, b, c):
    """
    (a, b, c) -> (a, a xor b, a.b xor c)

    Args:
        a(Qubit): qubit
        b(Qubit): qubit
        c(Qubit): qubit
    """
    with gate_set:
        CCX & (a, b, c)
        CX & (a, b)


def adder_overflow(gate_set, a, b, overflow):
    """
     store a + b in b

    (a,b,overflow) -> (a,b'=a+b,overflow'=overflow xor highest_carry)

    Args:
        a(Qureg): the qureg stores a, length is n
        b(Qureg): the qureg stores b, length is n
        overflow(Qureg): the ancillary qubits, length is 1

    reference: HIMANSHU THAPLIYAL and NAGARAJAN RANGANATHAN -
    Design of Efficient Reversible Logic Based Binary and BCD Adder Circuits
    https://arxiv.org/abs/1712.02630v1
    """

    n = len(a)
    with gate_set:
        if n == 1:
            peres_gate(gate_set, a[0], b[0], overflow)
            return

        # step 1
        for i in range(n - 1):
            CX & (a[i], b[i])

        # step 2
        CX & (a[0], overflow)
        for i in range(n - 2):
            CX & (a[i + 1], a[i])

        # step 3
        for i in range(n - 1):
            CCX & (a[n - 1 - i], b[n - 1 - i], a[n - 2 - i])

        # step 4
        peres_gate(gate_set, a[0], b[0], overflow)
        for i in range(n - 1):
            peres_gate(gate_set, a[i + 1], b[i + 1], a[i])

        # step 5
        for i in range(n - 2):
            CX & (a[n - 2 - i], a[n - 3 - i])

        # step 6
        for i in range(n - 1):
            CX & (a[i], b[i])


def adder(gate_set, a, b):
    """
    (a,b) -> (a,b'=a+b)

    Args:
        a(Qureg): the qureg stores a, length is n
        b(Qureg): the qureg stores b, length is n

    reference: HIMANSHU THAPLIYAL and NAGARAJAN RANGANATHAN -
    Design of Efficient Reversible Logic Based Binary and BCD Adder Circuits
    https://arxiv.org/abs/1712.02630v1
    """

    n = len(a)

    with gate_set:
        # step 1
        for i in range(n - 1):
            CX & (a[i], b[i])

        # step 2
        for i in range(n - 2):
            CX & (a[i + 1], a[i])

        # step 3
        for i in range(n - 1):
            CCX & (a[n - 1 - i], b[n - 1 - i], a[n - 2 - i])

        # step 4
        CX & (a[0], b[0])
        for i in range(n - 1):
            peres_gate(gate_set, a[i + 1], b[i + 1], a[i])

        # step 5
        for i in range(n - 2):
            CX & (a[n - 2 - i], a[n - 3 - i])

        # step 6
        for i in range(n - 1):
            CX & (a[i], b[i])


def subtraction_overflow(gate_set, a, b, overflow):
    """
    (a,b) -> (a,b-a)
    """
    n = len(b)
    with gate_set:
        for i in range(n):
            X & b[i]
        X & overflow
        adder_overflow(gate_set, a, b, overflow)
        for i in range(n):
            X & b[i]
        X & overflow


def ctrl_add_overflow_ancilla(gate_set, ctrl, a, b, overflow, ancilla):
    """
    (c,a,b,of,ancilla) -> (c,a,b+c*a,of',ancilla)
    """
    n = len(a)
    with gate_set:
        if n == 1:
            # CCX | (a[0],b[0],overflow)
            CCX & (a[0], b[0], ancilla)
            CCX & (ancilla, ctrl, overflow)
            CCX & (a[0], b[0], ancilla)
            CX  & (a[0], b[0])
            return

        # step 1
        for i in range(n - 1):
            CX & (a[i], b[i])
        # step 2
        CCX & (ctrl, a[0], overflow)
        for i in range(n - 2):
            CX & (a[i + 1], a[i])
        # step 3
        for i in range(n - 1):
            CCX & (a[n - 1 - i], b[n - 1 - i], a[n - 2 - i])
        # step 4
        CCX & (a[0], b[0], ancilla)
        CCX & (ctrl, ancilla, overflow)
        CCX & (a[0], b[0], ancilla)
        CCX & (ctrl, a[0], b[0])
        # step 5
        for i in range(n - 1):
            CCX & (a[i + 1], b[i + 1], a[i])
            CCX & (ctrl, a[i + 1], b[i + 1])
        # step 6
        for i in range(n - 2):
            CX & (a[n - 2 - i], a[n - 3 - i])
        # step 7
        for i in range(n - 1):
            CX & (a[i], b[i])


def ctrl_add(gate_set, ctrl, a, b):
    """
    (ctrl,a,b) -> (ctrl,a,b+a)
    """
    n = len(a)
    # step 1
    with gate_set:
        for i in range(n - 1):
            CX & (a[i], b[i])
        # step 2
        # CCX | (ctrl,a[0],ancilla)
        for i in range(n - 2):
            CX & (a[i + 1], a[i])
        # step 3
        for i in range(n - 1):
            CCX & (a[n - 1 - i], b[n - 1 - i], a[n - 2 - i])
        # step 4
        # CCX | (a[0],b[0],ancilla)
        # CCX | (ctrl,ancilla,overflow)
        # CCX | (a[0],b[0],ancilla)
        CCX & (ctrl, a[0], b[0])
        # step 5
        for i in range(n - 1):
            CCX & (a[i + 1], b[i + 1], a[i])
            CCX & (ctrl, a[i + 1], b[i + 1])
        # step 6
        for i in range(n - 2):
            CX & (a[n - 2 - i], a[n - 3 - i])
        # step 7
        for i in range(n - 1):
            CX & (a[i], b[i])


def mult(gate_set, a, b, p, ancilla):
    """
    Multiplicant: a, b
    Product: p
    (a, b, p=0, ancilla=0) -> (a, b, a*b, ancilla=0)
    """
    n = len(a)
    with gate_set:
        if n == 1:
            CCX & (a[0], b[0], p[1])
            return

        # step 1
        for i in range(n):
            CCX & (b[n - 1], a[n - 1 - i], p[2 * n - 1 - i])
        # step 2
        for i in range(n - 2):
            ctrl_add_overflow_ancilla(
                gate_set,
                b[n - 2 - i],
                a,
                p[n - 1 - i:2 * n - 1 - i],
                p[n - 2 - i],
                p[n - 3 - i]
            )
        ctrl_add_overflow_ancilla(
            gate_set,
            b[0],
            a,
            p[1:n + 1],
            p[0],
            ancilla
        )


def division(gate_set, a, b, r, ancilla):
    """
    Divided: a
    Divisor: b
    (a,b,r=0,ancilla=0) -> (a%b,b,a//b,ancilla)
    """
    n = len(a)
    with gate_set:
        for i in range(n - 1):
            # Iteration(y,b,r[i])
            y = r[i + 1:n] + a[0:i + 1]
            subtraction_overflow(gate_set, b, y, ancilla)
            CX & (ancilla, r[i])
            ctrl_add(gate_set, ancilla, b, y)
            CX & (r[i], ancilla)
            X & r[i]
        # Iteration(a,b,r[n-1])
        subtraction_overflow(gate_set, b, a, ancilla)
        CX & (ancilla, r[n - 1])
        ctrl_add(gate_set, ancilla, b, a)
        CX & (r[n - 1], ancilla)
        X & r[n - 1]


class RippleCarryAdder(object):
    """
    (a,b) -> (a,b'=a+b)
    Args:
        n(int): the bit number of a and b

    Quregs:
        a_q(Qureg): the qureg stores a, n qubits.
        b_q(Qureg): the qureg stores b, and stores the sum
                    after computation, n qubits.
    reference: HIMANSHU THAPLIYAL and NAGARAJAN RANGANATHAN -
    Design of Efficient Reversible Logic Based Binary and BCD adder Circuits
    https://arxiv.org/abs/1712.02630v1
    """

    @staticmethod
    def execute(n):
        """
        Construct a compositegate tailored to the size n
        """
        gate_set = CompositeGate()
        a_q = list(range(n))
        b_q = list(range(n, 2 * n))

        adder(gate_set, a_q, b_q)
        return gate_set


class Multiplication(object):
    """
    (a,b,p=0,ancilla=0) -> (a,b,p=a*b,ancilla=0)

    Args:
        n(int): the bit number of a and b

    Quregs:
        a_q(Qureg): the qureg stores a, n qubits.
        b_q(Qureg): the qureg stores b, n qubits.
        p_q(Qureg): the qureg stores the product, 2n qubits.
        ancilla(Qubit): the clean ancilla qubit, 1 qubit.

    reference: Edgard Mu˜noz-Coreas, Himanshu Thapliya -
    T-count Optimized Design of Quantum Integer Multiplication
    https://arxiv.org/abs/1706.05113v1
    """

    @staticmethod
    def execute(n):
        """
        Construct a compositegate tailored to the size n
        """
        gate_set = CompositeGate()
        a_q = list(range(n))
        b_q = list(range(n, 2 * n))
        p_q = list(range(2 * n, 4 * n))
        ancilla = 4 * n

        mult(gate_set, a_q, b_q, p_q, ancilla)

        return gate_set


class RestoringDivision(object):
    """
    (a,b,r=0,overflow=0) -> (a%b,b,a//b,0)
    Args:
        n(int): the bit number of a and b
    Quregs:
        a_q(Qureg): the qureg stores a, n qubits.
        b_q(Qureg): the qureg stores b, and stores the quotient
                    after computation, n qubits.
        r_q(Qureg): the qureg stores the remainder, n qubits.
        of_q(Qubit): the clean ancilla qubit, 1 qubit.
    reference: Himanshu Thapliyal, Edgard Munoz-Coreas, T. S. S. Varun, and Travis S. Humble -
    Quantum Circuit Designs of Integer Division Optimizing T-count and T-depth
    http://arxiv.org/abs/1809.09732v1
    """

    @staticmethod
    def execute(n):
        """
        Construct a compositegate tailored to the size n
        """
        gate_set = CompositeGate()
        a_q = list(range(n))
        b_q = list(range(n, 2 * n))
        r_q = list(range(2 * n, 3 * n))
        of_q = 3 * n

        division(gate_set, a_q, b_q, r_q, of_q)

        return gate_set



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/14 10:54
# @Author  : Han Yu, Li Kaiqi
# @File    : gate_builder.py

import numpy as np
import random

from QuICT.core.qubit import Qubit, Qureg
from QuICT.core.gate import *
from QuICT.core.utils import GateType
from QuICT.tools.exception.core import GateQubitAssignedError


GATE_TYPE_TO_CLASS = {
    GateType.h: HGate,
    GateType.hy: HYGate,
    GateType.s: SGate,
    GateType.sdg: SDaggerGate,
    GateType.x: XGate,
    GateType.y: YGate,
    GateType.z: ZGate,
    GateType.sx: SXGate,
    GateType.sy: SYGate,
    GateType.sw: SWGate,
    GateType.id: IDGate,
    GateType.u1: U1Gate,
    GateType.u2: U2Gate,
    GateType.u3: U3Gate,
    GateType.rx: RxGate,
    GateType.ry: RyGate,
    GateType.rz: RzGate,
    GateType.t: TGate,
    GateType.tdg: TDaggerGate,
    GateType.phase: PhaseGate,
    GateType.gphase: GlobalPhaseGate,
    GateType.cz: CZGate,
    GateType.cx: CXGate,
    GateType.cy: CYGate,
    GateType.ch: CHGate,
    GateType.crz: CRzGate,
    GateType.cu1: CU1Gate,
    GateType.cu3: CU3Gate,
    GateType.fsim: FSimGate,
    GateType.rxx: RxxGate,
    GateType.ryy: RyyGate,
    GateType.rzz: RzzGate,
    GateType.rzx: RzxGate,
    GateType.swap: SwapGate,
    GateType.iswap: iSwapGate,
    GateType.iswapdg: iSwapDaggerGate,
    GateType.sqiswap: SquareRootiSwapGate,
    GateType.cswap: CSwapGate,
    GateType.ccx: CCXGate,
    GateType.ccz: CCZGate,
    GateType.ccrz: CCRzGate,
    GateType.measure: MeasureGate,
    GateType.reset: ResetGate,
    GateType.barrier: BarrierGate,
    GateType.unitary: UnitaryGate,
    GateType.perm: PermGate,
    GateType.perm_fx: PermFxGate,
    GateType.qft: QFTGate,
    GateType.iqft: IQFTGate
}


def build_gate(
    gate_type: GateType,
    qubits: list,
    params: list = None
):
    """ Build a quantum gate with given parameter

    Args:
        gate_type (GateType): The gate type
        qubits (int/list/Qubit/Qureg): The qubit or qubit index of the gate.
        params (list, optional): The gate's parameters. Defaults to None.

    Returns:
        BasicGate: The quantum gate
    """
    gate = GATE_TYPE_TO_CLASS[gate_type]()
    if params is not None:
        params = params if isinstance(params, list) else [params]
        gate = gate(*params)

    args_number = gate.controls + gate.targets
    if isinstance(qubits, Qubit):
        qubits = Qureg(qubits)
    elif isinstance(qubits, int):
        qubits = [qubits]
    assert len(qubits) == args_number, GateQubitAssignedError(
        "The qubits number should equal to the target quantum gate."
    )

    if isinstance(qubits, Qureg):
        gate.assigned_qubits = qubits
    else:
        gate.cargs = qubits[:gate.controls]
        gate.targs = qubits[gate.controls:]

    return gate


def build_random_gate(
    gate_type: GateType,
    qubits: int,
    random_params: bool = False
):
    """ Build a quantum gate with random qubit and parameter.

    Args:
        gate_type (GateType): The gate type
        qubits (int): the number of qubits
        random_params (bool, optional): whether random parameter or use default parameter. Defaults to False.

    Returns:
        BasicGate: The quantum gate
    """
    gate = GATE_TYPE_TO_CLASS[gate_type]()
    args_number = gate.controls + gate.targets
    choiced_qubits = random.sample(range(qubits), args_number)
    gate.cargs = choiced_qubits[:gate.controls]
    gate.targs = choiced_qubits[gate.controls:]

    if random_params and gate.params:
        gate.pargs = list(np.random.uniform(0, 2 * np.pi, gate.params))

    return gate



import random
from math import gcd
import numpy as np
from fractions import Fraction
from typing import List, Tuple

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.simulation.state_vector import CircuitSimulator
from .utility import *

from .BEA_zip import construct_circuit as BEA_zip_circuit
from .BEA_zip import order_finding as BEA_zip_run
from .BEA import construct_circuit as BEA_circuit
from .BEA import order_finding as BEA_run

from .HRS_zip import construct_circuit as HRS_zip_circuit
from .HRS_zip import order_finding as HRS_zip_run
from .HRS import construct_circuit as HRS_circuit
from .HRS import order_finding as HRS_run

from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("Shor")


class ShorFactor:

    _ALLOWED_MODES = {"BEA", "HRS", "BEA_zip", "HRS_zip"}
    _RUN_METHOD_OF_MODE = {
        "BEA": reinforced_order_finding_constructor(BEA_run),
        "HRS": reinforced_order_finding_constructor(HRS_run),
        "BEA_zip": reinforced_order_finding_constructor(BEA_zip_run),
        "HRS_zip": reinforced_order_finding_constructor(HRS_zip_run),
    }
    _CIRCUIT_METHOD_OF_MODE = {
        "BEA": BEA_circuit,
        "HRS": HRS_circuit,
        "BEA_zip": BEA_zip_circuit,
        "HRS_zip": HRS_zip_circuit,
    }

    # add a, N here
    def __init__(self, mode: str, eps: float = 1 / 10, max_rd: int = 2, simulator=CircuitSimulator()) -> None:
        if mode not in ShorFactor._ALLOWED_MODES:
            raise ValueError(
                f"{mode} mode is not valid. Consider {ShorFactor._ALLOWED_MODES}"
            )
        random.seed(2022)
        self.mode = mode
        self.eps = eps
        self.max_rd = max_rd
        self.simulator = simulator
        self._previous_N = None

    def circuit(self, N: int) -> Tuple[Circuit, List[int]]:
        """construct the quantum part of Shor algorithm, i.e. order finding circuit

        Args:
            N (int): number to be factored

        Returns:
            Circuit: order finding circuit that can be passed to ShorFactor::run method
            List[int]: the indices to be measured to get ~phi
        """
        if self._previous_N is not None and self._previous_N == N:
            self._circuit_cache[0].reset_qubits()
            return self._circuit_cache
        # without usable previous circuit
        a = N
        while gcd(a, N) != 1:
            a = random.randrange(0, N)
        self._circuit_cache = ShorFactor._CIRCUIT_METHOD_OF_MODE[self.mode](a, N, self.eps)
        return self._circuit_cache

    def run(
        self,
        N,
        circuit: Circuit = None,
        indices: List = None,
        forced_quantum_approach=False
    ) -> int:
        """run full factoring algorithm.

        Args:
            N (int): number to be factored
            circuit (Circuit, optional): if None, a circuit will be constructed for order finding purpose;
                else the given circuit is used as order finding circuit. Defaults to None.
            indices (List, optional): The indices of $\\ket{n}$. Only used when `circuit` is not None. Defaults to None.
            forced_quantum_approach (bool, optional): If true, only x that gcd(x,N)=1 will be used. Defaults to False.

        Returns:
            int: the factor
        """
        simulator = self.simulator
        # check if input is prime (using MillerRabin in klog(N), k is the number of rounds to run MillerRabin)
        if miller_rabin(N):
            logger.info("N does not pass miller rabin test, may be a prime number")
            return 0
        # 1. If n is even, return the factor 2
        if N % 2 == 0:
            logger.info("Shor succeed: N is even, found factor 2 classically")
            return 2
        # 2. Classically determine if N = p^q
        y, L = np.log2(N), int(np.ceil(np.log2(N)))
        for b in range(2, L):
            squeeze = np.power(2, y / b)
            u1, u2 = int(np.floor(squeeze)), int(np.ceil(squeeze))
            if pow(u1, b) == N:
                logger.info(
                    f"Shor succeed: N is exponential, found the only factor {u1} classically"
                )
                return u1
            if pow(u2, b) == N:
                logger.info(
                    f"Shor succeed: N is exponential, found the only factor {u2} classically"
                )
                return u2
        rd = 0
        while rd < self.max_rd:
            logger.info(f"round = {rd}")
            # 3. Choose a random number a (1<a<N)
            if forced_quantum_approach:
                logger.info("forced quantum approach, looking for coprime number...")
                while True:
                    a = random.randint(2, N - 1)
                    gcd = np.gcd(a, N)
                    if gcd == 1:
                        break
            else:
                a = random.randint(2, N - 1)
                gcd = np.gcd(a, N)
                if gcd > 1:
                    logger.info(
                        f"Shor succeed: randomly chosen a = {a}, who has common factor {gcd} with N classically"
                    )
                    return gcd
            rd += 1
            # 4. Use quantum order-finding algorithm to find the order of a
            logger.info(
                f"Quantumly determine the order of the randomly chosen a = {a}"
            )
            # check if any input circuit. if no, run according to `mode`; else run the input circuit
            if circuit is None:
                r = ShorFactor._RUN_METHOD_OF_MODE[self.mode](
                    a=a, N=N, simulator=simulator
                )
            else:
                circuit.reset_qubits()
                simulator.run(circuit)
                if len(indices) > 0 and self.mode in {"BEA", "HRS"}:
                    phi = int(circuit[indices]) / (1 << len(indices))
                elif len(indices) > 0 and self.mode in {"BEA_zip", "HRS_zip"}:
                    phi = eval(
                        "0b" + "".join([str(circuit[indices[0]].historical_measured[idx]) for idx in indices[1:]])
                    ) / (1 << len(indices))
                else:
                    raise ValueError("wrong indices")
                logger.info(f"phi: {phi:4.3f}")
                r = Fraction(phi).limit_denominator(N - 1).denominator

            if r == 0:  # no order found
                logger.info(
                    f"Shor failed: did not find the order of a = {a}"
                )  # add mode info here
            elif r % 2 == 1:  # odd order
                logger.info(f"Shor failed: found odd order r = {r} of a = {a}")
            else:  # N | h^2 - 1, h = a^(r/2)
                h = pow(a, int(r / 2), N)
                if h == N - 1:  # N | h + 1
                    logger.info(
                        f"Shor failed: found order r = {r} of a = {a} with negative square root"
                    )
                else:  # N !| h + 1, therefore N | h - 1
                    f1, f2 = np.gcd(h - 1, N), np.gcd(h + 1, N)
                    if f1 > 1 and f1 < N:
                        logger.info(
                            f"Shor succeed: found factor {f1}, with the help of a = {a}, r = {r}"
                        )
                        return f1
                    elif f2 > 1 and f2 < N:
                        logger.info(
                            f"Shor succeed: found factor {f2}, with the help of a = {a}, r = {r}"
                        )
                        return f2
                    else:
                        logger.info(f"Shor failed: can not find a factor with a = {a}")
        return 0



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/5/27 下午4:23
# @Author  : Kaiqi Li
# @File    : proxy

import cupy as cp
import numpy as np
from cupy import cuda
from cupy.cuda import nccl

from typing import Union
from QuICT.utility.timeout import timeout


# Mapping between dtype and nccl dtype
# Modify for cupy>9.0.0

type_mapping = {
    "int8": np.int8,
    "uint8": np.uint8,
    "int32": np.int32,
    "uint32": np.uint32,
    "int64": np.int64,
    "uint64": np.uint64,
    "float16": np.float16,
    "float32": np.float32,
    "float64": np.float64,
    "complex64": np.complex64,
    "complex128": np.complex128
}


TIMEOUT = 300   # The timeout for receiving data


class Proxy:
    """ The proxy class of NCCL Communicators, which is used to transfer data between gpus.

    Args:
        ndevs(int): number of total GPU devices.
        uid(tuple): The unique ID, generate by cupy.cuda.nccl.get_unique_id().
        dev_id(int): The dev_id of GPU, between 0 and ndevs-1; represent the gpu device use in current process.
        timeout(int): The maximum waiting time for receiving data, default to 300 seconds.
    """
    def __init__(self, ndevs: int, uid: tuple, dev_id: int, timeout: int = 300):
        assert((dev_id < ndevs) and (ndevs <= cuda.runtime.getDeviceCount()))
        if dev_id != cuda.runtime.getDevice():
            target_device = cuda.Device(dev_id)
            target_device.use()

        self._ndevs = ndevs
        self._uid = uid
        self._dev_id = dev_id
        self._timeout = timeout
        self.peers = np.setdiff1d(np.arange(self._ndevs), self._dev_id)

        self.comm = nccl.NcclCommunicator(self._ndevs, self._uid, self._dev_id)

    def __exit__(self):
        self.comm.destroy()

    @staticmethod
    def allInit(devs: Union[int, list]):
        """ Initialize NCCL communicators for multiple devices in a single process. """
        return nccl.NcclCommunicator.initAll(devs)

    @property
    def ndevs(self):
        """ return the number of GPU devices. """
        return self._ndevs

    @property
    def uid(self):
        """ return the unique id of current communicator. """
        return self._uid

    @property
    def dev_id(self):
        """ return the dev_id of current communicator. """
        return self._dev_id

    def send(
        self,
        sendbuf: cp.ndarray,
        destination: int,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Send data to destination.

        Args:
            sendbuf(cupy.ndarray): the sending data.
            destination(int): the dev_id of destination communicator.
            stream(Stream): the cupy stream.
        """
        assert(type(sendbuf) == cp.ndarray)
        pointer = sendbuf.data.ptr
        count = sendbuf.size
        nccl_datatype = type_mapping[str(sendbuf.dtype)]

        if sendbuf.dtype == cp.complex128:
            count *= 2

        self.comm.send(pointer, count, nccl_datatype, destination, stream)

    @timeout()
    def recv(
        self,
        recvbuf: cp.ndarray,
        source: int,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Receive data from source and put it into recvbuf.

        Args:
            recvbuf(cupy.ndarray): the GPU array waitting for comming data.
            source(int): the dev_id of the source communicator.
            stream(Stream): the cupy stream.
        """
        count = recvbuf.size
        nccl_datatype = type_mapping[str(recvbuf.dtype)]

        if recvbuf.dtype == cp.complex128:
            count *= 2

        self.comm.recv(recvbuf.data.ptr, count, nccl_datatype, source, stream)

        return recvbuf

    def broadcast(
        self,
        databuf: cp.ndarray,
        root: int = 0,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Send data to all communicators except root. The sending data is in sendbuf of root communicator.

        Args:
            databuf(cupy.ndarray): the sending data in root, and the received array for other communicators.
            root(int): the dev_id ID of root communicator.
            stream(Stream): the cupy stream.
        """
        pointer = databuf.data.ptr
        count = databuf.size
        nccl_datatype = type_mapping[str(databuf.dtype)]

        if databuf.dtype == cp.complex128:
            count *= 2

        self.comm.bcast(pointer, count, nccl_datatype, root, stream)

    def reduce(
        self,
        root: int,
        op: int,
        sendbuf: cp.ndarray = None,
        recvbuf: cp.ndarray = None,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Compute the op with the data from other peers, and write the result in the root.

        Args:
            root(int): the dev_id ID of root communicator.
            op (int): The operation with one of
                NCCL_SUM = 0,
                NCCL_PROD = 1,
                NCCL_MAX = 2,
                NCCL_MIN = 3.
            sendbuf(cupy.ndarray): the sending data, not used in the root.
            recvbuf(cupy.ndarray): the array waitting for comming data, must be defined in the root.
            stream(Stream): the cupy stream.
        """
        if self._dev_id == root:
            assert(recvbuf is not None)
            sendbuf = cp.empty(1, dtype=recvbuf.dtype)
            pointer = recvbuf.data.ptr
            count = recvbuf.size
            nccl_datatype = type_mapping[str(recvbuf.dtype)]

            if recvbuf.dtype == cp.complex128:
                count *= 2

            self.comm.reduce(sendbuf.data.ptr, pointer, count, nccl_datatype, op, root, stream)
        else:
            assert(sendbuf is not None)
            recvbuf = cp.empty(1, dtype=sendbuf.dtype)
            pointer = sendbuf.data.ptr
            count = sendbuf.size
            nccl_datatype = type_mapping[str(sendbuf.dtype)]

            if sendbuf.dtype == cp.complex128:
                count *= 2

            self.comm.reduce(pointer, recvbuf.data.ptr, count, nccl_datatype, op, root, stream)

    def reducescatter(
        self,
        sendbuf: cp.ndarray,
        recvbuf: cp.ndarray,
        op: int,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Compute the op with the data from all peers, and split the result into all peers.

        Args:
            sendbuf(cupy.ndarray): the sending data.
            recvbuf(cupy.ndarray): the array waitting for the result.
            op (int): The operation with one of
                NCCL_SUM = 0,
                NCCL_PROD = 1,
                NCCL_MAX = 2,
                NCCL_MIN = 3.
            stream(Stream): the cupy stream.
        """
        send_pointer = sendbuf.data.ptr
        recv_pointer = recvbuf.data.ptr
        count = recvbuf.size
        nccl_datatype = type_mapping[str(sendbuf.dtype)]

        if sendbuf.dtype == cp.complex128:
            count *= 2

        self.comm.reduceScatter(send_pointer, recv_pointer, count, nccl_datatype, op, stream)

    def allreduce(
        self,
        sendbuf: cp.ndarray,
        recvbuf: cp.ndarray,
        op: int,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Compute the op with the data from all peers, and send the result to all peers.

        Args:
            sendbuf(cupy.ndarray): the sending data.
            recvbuf(cupy.ndarray): the array waitting for the result.
            op (int): The operation with one of
                NCCL_SUM = 0,
                NCCL_PROD = 1,
                NCCL_MAX = 2,
                NCCL_MIN = 3.
            stream(Stream): the cupy stream.
        """
        send_pointer = sendbuf.data.ptr
        recv_pointer = recvbuf.data.ptr
        count = sendbuf.size
        nccl_datatype = type_mapping[str(sendbuf.dtype)]

        if sendbuf.dtype == cp.complex128:
            count *= 2

        self.comm.allReduce(send_pointer, recv_pointer, count, nccl_datatype, op, stream)

    def scatter(
        self,
        sendbuf: cp.ndarray,
        targets: list,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Split sending data, and send to all peers in targets.
        Note that target proxy should using proxy.recv to receive the piece of sending data.

        Args:
            sendbuf(cupy.ndarray): the sending data.
            targets(list): List of the communicators' dev_id ID which hope to receive.
            stream(Stream): the cupy stream.
        """
        # Divided data depending on the given rules
        dest_len = len(targets)
        data_interval = sendbuf.size // dest_len

        # Send data to each target
        for dest_idx, dest in enumerate(targets):
            self.send(sendbuf[dest_idx * data_interval:(dest_idx + 1) * data_interval], dest, stream)

    def gather(
        self,
        root: int,
        sendbuf: cp.ndarray = None,
        recvbuf: cp.ndarray = None,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Collect all data from peers into the root.

        Args:
            root(int): the dev_id ID of the root communicator.
            sendbuf(cupy.ndarray): the sending data, not use in the root communicator.
            recvbuf(cupy.ndarray): the GPU array waitting for comming data, only use in the
            root communicator.
            stream(Stream): the cupy stream.
        """
        if self._dev_id == root:
            assert(recvbuf is not None)

            recv_count_per_dev = recvbuf.size // (self._ndevs - 1)
            for dest in self.peers:
                self.recv(recvbuf[(dest - 1) * recv_count_per_dev:dest * recv_count_per_dev], dest, stream)
        else:
            assert(sendbuf is not None)

            self.send(sendbuf, root, stream)

    def allgather(
        self,
        sendbuf: cp.ndarray,
        recvbuf: cp.ndarray,
        stream: cuda.stream.Stream = cuda.Stream.null.ptr
    ):
        """
        Collect all data from peers, and send the aggregation bact to each peer.

        Args:
            sendbuf(cupy.ndarray): the sending data.
            recvbuf(cupy.ndarray): the GPU array waitting for comming data.
            stream(Stream): the cupy stream.
        """
        send_pointer = sendbuf.data.ptr
        recv_pointer = recvbuf.data.ptr
        count = sendbuf.size
        nccl_datatype = type_mapping[str(sendbuf.dtype)]

        if sendbuf.dtype == cp.complex128:
            count *= 2

        self.comm.allGather(send_pointer, recv_pointer, count, nccl_datatype, stream)



from typing import Union, Dict

from QuICT.tools import Logger
from QuICT.tools.logger import LogFormat
from QuICT.tools.cli.utils import JobValidation
from .utils import EncryptedRequest, EncryptManager, SQLManager


# TODO: file copy between local and remote
default_hostname = "127.0.0.1"
default_api_server_port = "5000"


def login_validation(function):
    def decorator(self, *args, **kwargs):
        login, err_msg = self._sql_db.login_validation()
        if not login:
            self._logger.warn(err_msg)
            return

        login_info = self._sql_db.get_login_info()

        result = function(self, user_info=login_info, *args, **kwargs)

        return result

    return decorator


class QuICTRemoteManager:
    """ QuICT Job Management for the Remote Mode. """
    def __init__(
        self,
        hostname: str = default_hostname,
        api_server_port: str = default_api_server_port
    ):
        self._url_prefix = f"http://{hostname}:{api_server_port}/quict"
        self._encrypt = EncryptManager()
        self._encryptedrequest = EncryptedRequest()
        self._sql_db = SQLManager()
        self._logger = Logger("Job_Management_Remote_Mode", LogFormat.full)

    ####################################################################
    ############                Login & Logout              ############
    ####################################################################
    def login(self, username: str, password: str):
        """ User login.

        Args:
            username (str): The username
            password (str): The password
        """
        encrypted_passwd = self._encrypt.encrypted_passwd(password)
        success = self._encryptedrequest.post(
            f"{self._url_prefix}/env/login",
            {'username': username, 'password': encrypted_passwd},
            (1, username, encrypted_passwd),
            is_login=True
        )

        if not success:
            self._logger.warn("unmatched login username and password.")
        else:
            self._sql_db.user_login(username, encrypted_passwd)
            self._logger.info("successfully login.")

    def logout(self):
        """ User logout. """
        self._sql_db.user_logout()
        self._logger.info("successfully logout.")

    def register(
        self, username: str, password: str, email: str
    ):
        """ User register.

        Args:
            username (str): The username
            password (str): The password
            email (str): The email address
            level (int): The difficult level
        """
        encrypted_passwd = self._encrypt.encrypted_passwd(password)
        success = self._encryptedrequest.post(
            f"{self._url_prefix}/env/register",
            {
                'username': username,
                'password': encrypted_passwd,
                'email': email
            },
            (1, username, encrypted_passwd),
            is_login=True
        )

        if not success:
            self._logger.warn("Failure to register with current username.")
        else:
            self._sql_db.user_login(username, encrypted_passwd)
            self._logger.info(f"Successfully register with username {username}.")

    def unsubscribe(self, username: str, password: str):
        """ delete user's account.

        Args:
            username (str): The username
            password (str): The password
        """
        encrypted_passwd = self._encrypt.encrypted_passwd(password)
        success = self._encryptedrequest.post(
            f"{self._url_prefix}/env/unsubscribe",
            {
                'username': username,
                'password': encrypted_passwd,
            },
            (1, username, encrypted_passwd),
            is_login=False
        )

        if not success:
            self._logger.warn("Failure to unsubscribe with current username.")
        else:
            self._logger.info(f"Successfully unsubscribe with username {username}.")

    ####################################################################
    ############               Job API Function             ############
    ####################################################################
    @login_validation
    def start_job(self, yml_dict: Union[Dict, str], user_info: tuple = None):
        """ Send job ticket to cloud system.

        Args:
            yml_dict (dict): The job's information
            user_info (dict): The user's information
        """
        # Job Dict Validation
        job_info = JobValidation().job_validation(yml_dict)

        # Delete Circuit Qasm Path here, not use for remote mode
        self._remote_job_validate(job_info)

        url = f"{self._url_prefix}/jobs/start"
        try:
            _ = self._encryptedrequest.post(url, job_info, user_info=user_info)
            self._logger.info("Successfully send job to cloud.")
        except Exception as e:
            self._logger.warn(f"Failure to start target job, due to {e}.")

    def _remote_job_validate(self, yml_dict: dict):
        # Deal with circuit, load circuit's qasm.
        with open(yml_dict['circuit']) as cfile:
            circuit_data = cfile.read()

        del yml_dict['circuit']
        yml_dict['circuit_info']['qasm'] = circuit_data

        # Deal with layout
        if "qcda" in yml_dict.keys() and "layout_path" in yml_dict["qcda"].keys():
            with open(yml_dict['qcda']['layout_path']) as lfile:
                layout_data = lfile.read()

            del yml_dict['qcda']['layout_path']
            yml_dict['qcda']['layout_string'] = layout_data

    @login_validation
    def status_job(self, job_name: str, user_info: tuple):
        """ Check a job's running states

        Args:
            job_name (str): The job's name
            user_info (tuple): The user's information
        """
        url = f"{self._url_prefix}/jobs/{job_name}:status"
        job_dict = self._encryptedrequest.get(url, user_info=user_info)
        job_name = job_dict["job_name"]
        job_status = job_dict["state"]
        self._logger.info(f"job name: {job_name}, job's state: {job_status}.")

    @login_validation
    def delete_job(self, job_name: str, user_info: tuple):
        """ Delete a job.

        Args:
            job_name (str): The job's name
            user_info (tuple): The user's information
        """
        url = f"{self._url_prefix}/jobs/{job_name}:delete"
        try:
            self._encryptedrequest.delete(url, user_info=user_info)
            self._logger.info("Successfully send delete request to cloud.")
        except Exception as e:
            self._logger.warn(f"Failure to delete target job, due to {e}.")

    @login_validation
    def list_jobs(self, user_info: tuple):
        """ List all jobs in the cloud system.

        Args:
            user_info (tuple): The user's information
        """
        url = f"{self._url_prefix}/jobs/list"
        job_list = self._encryptedrequest.get(url, user_info=user_info)
        self._logger.info(f"There are {len(job_list)} jobs in local mode.")
        for job in job_list:
            job_name = job["job_name"]
            job_status = job["state"]
            self._logger.info(f"job name: {job_name}, job's state: {job_status}.")



from flask import Blueprint
import json

from script.requset_validation import request_validation
from script.redis_controller import RedisController
from utils.data_structure import JobOperatorType
from utils.file_manage import create_job_folder


job_blueprint = Blueprint(name="jobs", import_name=__name__)
URL_PREFIX = "/quict/jobs"


@job_blueprint.route(f"{URL_PREFIX}/start", methods=["POST"])
@request_validation()
def start_job(**kwargs):
    """start a job. """
    job_dict = kwargs['json_dict']
    job_dict['username'] = kwargs['username']

    folder_path = create_job_folder(kwargs['username'], job_dict['job_name'])
    circuit_info = json.loads(job_dict['circuit_info'])
    # Create circuit's qasm and layout's json file if necessary
    with open(f"{folder_path}/circuit.qasm", 'w') as cw:
        cw.write(circuit_info['qasm'])

    del circuit_info['qasm']
    job_dict['circuit_info'] = json.dumps(circuit_info)

    if "qcda" in job_dict.keys():
        qcda_dict = json.loads(job_dict["qcda"])
        if "layout_string" in qcda_dict.keys():
            with open(f"{folder_path}/layout.json", 'w') as lw:
                lw.write(qcda_dict["layout_string"])

            del qcda_dict["layout_string"]

        job_dict["qcda"] = json.dumps(qcda_dict)

    # start job by redis controller
    RedisController().add_pending_job(job_dict)

    return True


@job_blueprint.route(f"{URL_PREFIX}/<name>:delete", methods=["DELETE"])
@request_validation()
def delete_job(name: str, username: str):
    """ delete a job. """
    RedisController().add_operator(f"{username}:{name}", JobOperatorType.delete)


@job_blueprint.route(f"{URL_PREFIX}/<name>:status", methods=["GET"])
@request_validation()
def status_jobs(name: str, username: str):
    return RedisController().get_job_info(f"{username}:{name}")


@job_blueprint.route(f"{URL_PREFIX}/list", methods=["GET"])
@request_validation()
def list_jobs(username: str):
    return RedisController().list_jobs(username)



import os
import yaml
# from kubenerte import client

from server.script.sql_controller import SQLManger


DEFAULT_YML_PATH = os.path.join(
    os.path.dirname(__file__),
    "../config"
)


def get_default_user_config(username: str):
    user_info_path = os.path.join(DEFAULT_YML_PATH, "user_info.yml")
    with open(user_info_path, encoding='utf-8') as yml:
        yaml_dict = yaml.load(yml)

    # Get user info from SQL database
    user_info = SQLManger().get_user_info(username)
    yaml_dict['username'] = username
    yaml_dict['maximum_parallel_level'] = user_info[3]
    yaml_dict['maximum_stop_level'] = user_info[4]
    yaml_dict['GPU_allowence'] = user_info[5]

    return yaml_dict


def get_default_job_config(job_info: dict):
    k8s_job_yml_path = os.path.join(DEFAULT_YML_PATH, "job.yml")
    with open(k8s_job_yml_path, encoding='utf-8') as yml:
        yaml_dict = yaml.load(yml)

    # get k8s job's metadata
    job_meta = {
        "name": job_info["job_name"],
        "userspace": job_info["username"]
    }
    yaml_dict["metadata"] = job_meta

    # Prepare start command
    container_related = yaml_dict["spec"]["template"]["spec"]["containers"][0]
    cmd = f"{job_info['type']}.py"
    container_related["command"].append(cmd)

    # Prepare environment args
    optional_args = job_info[job_info['type']]
    for key, value in optional_args.items():
        env_dict = {
            "name": key,
            "value": value
        }
        container_related["env"].append(env_dict)

    # Prepare MountPath
    mount_path = os.path.join(
        "/home/likaiqi/Cluster/User", job_info['username'], job_info['job_name']
    )
    volumes_related = yaml_dict["spec"]["template"]["spec"]['volumes'][0]
    volumes_related['hostPath']['path'] = mount_path

    return yaml_dict



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/12/28 12:29 上午
# @Author  : Han Yu
# @File    : uniformly_unitary.py

import numpy as np

from .uniformly_rotation import UniformlyRotation
from QuICT.core.gate import build_gate, GateType, CompositeGate, H, Rz, CX


class UniformlyUnitary(object):
    """
    Implements the uniformly unitary gate

    Reference:
        https://arxiv.org/abs/quant-ph/0504100 Fig4 b)
    """
    def execute(self, matrices):
        """
        Args:
            matrices(list<numpy.array>): the matrices of unitary gates

        Returns:
            CompositeGate: CompositeGate that implements the uniformly gate
        """
        matrices = list(matrices)
        n = int(np.log2(len(matrices))) + 1
        if 1 << (n - 1) != len(matrices):
            raise Exception("the number of parameters unmatched.")
        return self.uniformly_unitary(0, n, matrices)

    def uniformly_unitary(self, low, high, unitary):
        """ synthesis uniformlyUnitary gate, bits range [low, high)

        Args:
            low(int): the left range low
            high(int): the right range high
            unitary(list<int>): the list of unitaries
        Returns:
            the synthesis result
        """
        if low + 1 == high:
            rot = self.unitary_to_u3gate(unitary[0], low)
            gates = CompositeGate()
            gates.append(rot)
            return gates
        length = len(unitary) // 2
        Rxv = []
        Rxu = []
        angle_list = [0] * 2 * length
        for i in range(length):
            v, u, angles = self.get_parameters_from_unitaries(unitary[i], unitary[i + length])
            Rxu.append(u)
            Rxv.append(v)
            dual_position = 0
            for j in range(high - low - 2):
                if (1 << j) & i:
                    dual_position += 1 << (high - low - 2 - j - 1)
            angle_list[dual_position] = angles[0]
            angle_list[dual_position + length] = angles[1]
        gates = self.uniformly_unitary(low + 1, high, Rxv)
        CX & [low, high - 1] | gates
        gates.extend(self.uniformly_unitary(low + 1, high, Rxu))
        URz = UniformlyRotation(GateType.rz)
        urz = URz.execute(angle_list)
        urz & list(range(high - 1, low - 1, -1))
        gates.extend(urz)
        return gates

    def unitary_to_u3gate(self, unitary, target):
        """ gates from a one-qubit unitary

        Args:
            unitary(np.ndarray): the unitary to be transformed
            target(int): the qubit gate acts on
        Returns:
            U3Gate: gate from the unitary
        """
        unitary = np.array(unitary).reshape(2, 2)
        z = np.exp(1j * np.angle(unitary[0, 0]))
        unitary = unitary / z

        theta = np.arccos(unitary[0, 0])
        sint = np.sin(theta)
        if abs(sint) >= 1e-6:
            lamda = np.angle(unitary[0, 1] / -sint)
            phi = np.angle(unitary[1, 0] / sint)
        else:
            lamda = 0
            phi = np.angle(unitary[1, 1] / np.cos(theta))
        gate = build_gate(GateType.u3, target, [theta * 2, phi, lamda])
        assert np.allclose(gate.matrix, unitary)
        return gate

    def u2_phase_angle(self, mat):
        """ express U(2) with SU(2) and phase

        exp(i * phi / 2) SU(2) = U(2)

        Args:
            mat(np.ndarray): U2 matrix

        Returns:
            float: phase angle
        """
        if abs(mat[0, 0]) < 1e-10:
            absX = -mat[1, 0] * mat[0, 1]
            phase = np.angle(absX) / 2
            mat /= np.exp(1j * phase)
        else:
            absX = mat[0, 0] * mat[1, 1]
            phase = np.angle(absX) / 2
            mat /= np.exp(1j * phase)

        return 2 * phase

    def get_parameters_from_unitaries(self, u1, u2):
        """ decomposition uniformly controlled one qubit unitaries

        (0><0) ⊗ u1 + (1><1) ⊗ u2

        Args:
            u1(np.ndarray): unitary with 0
            u2(np.ndarray): unitary with 1

        Returns:
            np.ndarray: v in the decomposition
            np.ndarray: u in the decomposition
            list<float>: angle list of Rz
        """
        a = np.array(u1).reshape(2, 2)
        b = np.array(u2).reshape(2, 2)

        mat = a.dot(b.T.conj())
        phi = self.u2_phase_angle(mat)
        x1 = np.angle(mat[0, 0])

        r11_angle = 1j / 2 * (-np.pi / 2 - phi / 2 - x1)
        r22_angle = 1j / 2 * (np.pi / 2 - phi / 2 + x1)
        r = np.diag([np.exp(r11_angle), np.exp(r22_angle)])

        rXr = r.dot(mat).dot(r) * np.exp(1j * phi / 2)
        lamda, hU = np.linalg.eig(rXr)
        if abs(abs(lamda[0] - 1j)) >= 1e-10:
            hU[:, [0, 1]] = hU[:, [1, 0]]

        u = hU.reshape(2, 2)
        d = np.diag([np.exp(1j * np.pi / 4), np.exp(-1j * np.pi / 4)])
        v = d.T.conj().dot(u.T.conj()).dot(r).dot(a)
        v = H.matrix.dot(v)
        v *= np.exp(-1j * np.pi / 4)
        u = u.dot(Rz(-np.pi / 2).matrix).dot(H.matrix)

        return v, u, [-np.pi - 1.0 * (x1 + phi / 2), 1.0 * (x1 - phi / 2)]
